"""Utilizes same boxes as test_line_box_clip"""

import numpy as np

from numba_celltree.algorithms import cyrus_beck_line_polygon_clip as line_clip
from numba_celltree.algorithms.cyrus_beck import cyrus_beck_line_polygon_clip
from numba_celltree.constants import Point


def ab(a, b, c):
    """Flip the result around to compare (a, b) with (b, a)"""
    return (a, c, b)


def test_line_box_clip():
    poly = np.array(
        [
            [1.0, 3.0],
            [4.0, 3.0],
            [4.0, 5.0],
            [1.0, 5.0],
        ]
    )

    a = Point(0.0, 0.0)
    b = Point(4.0, 6.0)
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [2.0, 3.0])
    assert np.allclose(d, [3.3333333333333, 5.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))

    a = Point(0.0, 0.1)
    b = Point(0.0, 0.1)
    intersects, c, d = line_clip(a, b, poly)
    assert not intersects
    assert np.isnan(c).all()
    assert np.isnan(d).all()
    assert line_clip(a, b, poly)[0] == line_clip(b, a, poly)[0]

    a = Point(0.0, 4.0)
    b = Point(5.0, 4.0)
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.0, 4.0])
    assert np.allclose(d, [4.0, 4.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))

    poly = np.array(
        [
            [0.0, 0.0],
            [2.0, 0.0],
            [2.0, 2.0],
            [0.0, 2.0],
        ]
    )
    a = Point(1.0, -3.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 2.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))

    b = Point(1.0, 1.0)
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 1.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))

    a = Point(1.0, 1.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.0, 1.0])
    assert np.allclose(d, [1.0, 2.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))


def test_line_box_clip_through_vertex():
    a = Point(x=2.0, y=2.0)
    b = Point(x=1.0, y=4.0)
    poly = np.array([[2.0, 4.0], [1.0, 4.0], [1.0, 3.0], [2.0, 3.0]])
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.5, 3.0])
    assert np.allclose(d, [1.0, 4.0])

    a = Point(x=2.0, y=0.0)
    b = Point(x=2.0, y=2.0)
    poly = np.array([[2.0, 3.0], [1.0, 3.0], [1.0, 2.0], [2.0, 2.0]])
    intersects, _, _ = line_clip(a, b, poly)
    assert not intersects

    poly = np.array([[3.0, 3.0], [2.0, 3.0], [2.0, 2.0], [3.0, 2.0]])
    intersects, c, d = line_clip(a, b, poly)
    assert not intersects


def test_line_triangle_clip():
    # Triangle
    a = Point(1.0, 1.0)
    b = Point(3.0, 1.0)
    poly = np.array(
        [
            [0.0, 0.5],
            [2.0, 0.0],
            [2.0, 2.0],
        ]
    )
    intersects, c, d = line_clip(a, b, poly)
    assert intersects
    assert np.allclose(c, [1.0, 1.0])
    assert np.allclose(d, [2.0, 1.0])
    assert line_clip(a, b, poly) == ab(*line_clip(b, a, poly))


def test_line_triangle_clip_degeneracies():
    poly = np.array(
        [
            [0.0, 0.0],
            [2.0, 0.0],
            [2.0, 2.0],
        ]
    )
    # Lower edge
    a = Point(0.0, 0.0)
    b = Point(2.0, 0.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]

    # Right edge
    a = Point(2.0, 0.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]

    # Diagonal edge
    a = Point(0.0, 0.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]

    a = Point(-1.0, -1.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]

    a = Point(-1.0, -1.0)
    b = Point(3.0, 3.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]

    a = Point(0.0, 0.0)
    b = Point(3.0, 3.0)
    assert line_clip(a, b, poly)[0]
    assert line_clip(b, a, poly)[0]


def test_line_box_degeneracies():
    nodes = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [2.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0],
            [2.0, 1.0],
            [0.0, 2.0],
            [1.0, 2.0],
            [2.0, 2.0],
        ]
    )
    faces = np.array(
        [
            [3, 4, 6, 7],
            [4, 5, 8, 7],
            [0, 1, 4, 3],
            [1, 2, 5, 4],
        ]
    )
    poly = nodes[faces[1]]
    a = Point(1.0, 1.0)
    b = Point(1.5, 1.25)
    succes, c, d = cyrus_beck_line_polygon_clip(a, b, poly)
    assert succes
    assert c == Point(1.0, 1.0)
    assert d == Point(1.5, 1.25)

    for i in [0, 2, 3]:
        poly = nodes[faces[i]]
        a = Point(1.0, 1.0)
        b = Point(1.5, 1.25)
        succes, c, d = cyrus_beck_line_polygon_clip(a, b, poly)
        assert not succes

    # This is a case where a is inside, but b is right on the vertex
    # point_in_poly(a, poly) == True
    # point_in_poly(b, poly) == False
    # This results in t0 == t1 in Cyrus-Beck. However, we can get the right
    # by setting t0 to 0.0 if a_inside, or t1 to 1.0 if b_inside.
    poly = nodes[faces[2]]
    a = Point(0.5, 0.75)
    b = Point(1.0, 1.0)
    succes, c, d = cyrus_beck_line_polygon_clip(a, b, poly)
    assert succes
    assert c == a
    assert d == b



"""
Slightly adapted from:
https://github.com/scivision/lineclipping-python-fortran

(MIT License)
"""

from typing import Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import Box, Point

INSIDE, LEFT, RIGHT, LOWER, UPPER = 0, 1, 2, 4, 8


@nb.njit(inline="always")
def get_clip(a: Point, box: Box):
    p = INSIDE  # default is inside

    # consider x
    if a.x < box.xmin:
        p |= LEFT
    elif a.x > box.xmax:
        p |= RIGHT

    # consider y
    if a.y < box.ymin:
        p |= LOWER  # bitwise OR
    elif a.y > box.ymax:
        p |= UPPER  # bitwise OR
    return p


@nb.njit(inline="never")
def cohen_sutherland_line_box_clip(a: Point, b: Point, box: Box) -> Tuple[Point, Point]:
    """
    Clips a line to a rectangular area.

    This implements the Cohen-Sutherland line clipping algorithm.  xmin,
    ymax, xmax and ymin denote the clipping area, into which the line
    defined by a.x, a.y (start point) and b.x, b.y (end point) will be
    clipped.

    If the line does not intersect with the rectangular clipping area,
    four None values will be returned as tuple. Otherwise a tuple of the
    clipped line points will be returned in the form (cx1, ca.y, cb.x, cb.y).
    """
    NO_INTERSECTION = False, Point(np.nan, np.nan), Point(np.nan, np.nan)
    dx = b.x - a.x
    dy = b.y - a.y
    if dx == 0.0 and dy == 0.0:
        return NO_INTERSECTION

    # check for trivially outside lines
    k1 = get_clip(a, box)
    k2 = get_clip(b, box)

    # examine non-trivially outside points
    # bitwise OR |
    while (k1 | k2) != INSIDE:
        # if both points are inside box (0000),
        # ACCEPT trivial whole line in box, exit.

        # if line trivially outside window, REJECT
        if (k1 & k2) != 0:  # bitwise AND &
            return NO_INTERSECTION

        # non-trivial case, at least one point outside window
        # this is not a bitwise or, it's the word "or"
        opt = k1 or k2  # take first non-zero point, short circuit logic
        if opt & UPPER:  # these are bitwise ANDS
            x = a.x + dx * (box.ymax - a.y) / dy
            y = box.ymax
        elif opt & LOWER:
            x = a.x + dx * (box.ymin - a.y) / dy
            y = box.ymin
        elif opt & RIGHT:
            y = a.y + dy * (box.xmax - a.x) / dx
            x = box.xmax
        elif opt & LEFT:
            y = a.y + dy * (box.xmin - a.x) / dx
            x = box.xmin
        else:
            raise RuntimeError("Undefined clipping state")

        if opt == k1:
            a = Point(x, y)
            k1 = get_clip(a, box)
        elif opt == k2:
            b = Point(x, y)
            k2 = get_clip(b, box)

    return True, a, b



from typing import List, Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import (
    FLOAT_MAX,
    FLOAT_MIN,
    INT_MAX,
    Bucket,
    BucketArray,
    BucketDType,
    FloatArray,
    IntArray,
    IntDType,
    Node,
    NodeArray,
    NodeDType,
)
from numba_celltree.geometry_utils import build_bboxes
from numba_celltree.utils import allocate_stack, pop, push


@nb.njit(inline="always")
def create_node(ptr: int, size: int, dim: bool) -> Node:
    return Node(-1, -1.0, -1.0, ptr, size, dim)


@nb.njit(inline="always")
def push_node(nodes: NodeArray, node: Node, index: int) -> int:
    """Push to the end of the array."""
    nodes[index]["child"] = node.child
    nodes[index]["Lmax"] = node.Lmax
    nodes[index]["Rmin"] = node.Rmin
    nodes[index]["ptr"] = node.ptr
    nodes[index]["size"] = node.size
    nodes[index]["dim"] = node.dim
    return index + 1


@nb.njit(inline="always")
def centroid_test(bucket: np.void, box: FloatArray, dim: int):
    """
    Test whether the centroid of the bounding box in the selected dimension falls
    within this bucket.
    """
    centroid = box[2 * dim] + 0.5 * (box[2 * dim + 1] - box[2 * dim])
    return (centroid >= bucket.Min) and (centroid < bucket.Max)


@nb.njit(inline="never", cache=True)
def stable_partition(
    bb_indices: IntArray,
    bb_coords: FloatArray,
    begin: int,
    end: int,
    bucket: BucketDType,
    dim: int,
) -> int:
    """
    Rearrange the elements in the range(begin, end), in such a way that all
    the elements for which a predicate returns True precede all those for which it
    returns False. The relative order in each group is maintained.
    In this case, the predicate is a `centroid_test`.

    Parameters
    ----------
    bb_indices: np.ndarray of ints
        Array to sort.
    bb_coords: np.ndarray of floats
        Coordinates of bounding boxes.
    begin, end: int
        Defines the range of arr in which to sort.
    bucket: np.void
        Element of BucketArray, contains data for a single bucket.
    dim: int
        Dimension number (0: x, 1: y, etc.)

    Returns
    -------
    current: int
        Points to the first element of the second group for which predicate is True.
    """
    # Allocates a temporary buffer, ands fill from front and back: O(N)
    # A swapping algorithm can be found here, O(N log(N)):
    # https://csjobinterview.wordpress.com/2012/03/30/array-stable-partition/
    # via: https://stackoverflow.com/questions/21554635/how-is-stable-partition-an-adaptive-algorithm
    temp = np.empty(end - begin, dtype=bb_indices.dtype)
    # TODO: add statically allocated work-array? Then use views for size?

    count_true = 0
    count_false = -1
    for i in bb_indices[begin:end]:
        box = bb_coords[i]
        if centroid_test(bucket, box, dim):
            temp[count_true] = i
            count_true += 1
        else:
            temp[count_false] = i
            count_false -= 1

    for i in range(count_true):
        bb_indices[begin + i] = temp[i]

    start_second = begin + count_true
    for i in range(-1 - count_false):
        bb_indices[start_second + i] = temp[-i - 1]

    return start_second


@nb.njit(inline="never", cache=True)
def sort_bbox_indices(
    bb_indices: IntArray,
    bb_coords: FloatArray,
    buckets: BucketArray,
    node: NodeDType,
    dim: int,
):
    current = node.ptr
    end = node.ptr + node.size

    b = buckets[0]
    buckets[0] = Bucket(b.Max, b.Min, b.Rmin, b.Lmax, node.ptr, b.size)

    i = 1
    while current != end:
        bucket = buckets[i - 1]
        current = stable_partition(bb_indices, bb_coords, current, end, bucket, dim)
        start = bucket.index

        b = buckets[i - 1]
        buckets[i - 1] = Bucket(b.Max, b.Min, b.Rmin, b.Lmax, b.index, current - start)

        if i < len(buckets):
            b = buckets[i]
            buckets[i] = Bucket(
                b.Max,
                b.Min,
                b.Rmin,
                b.Lmax,
                buckets[i - 1].index + buckets[i - 1].size,
                b.size,
            )

        start = current
        i += 1


@nb.njit(inline="never", cache=True)
def get_bounds(
    index: int,
    size: int,
    bb_coords: FloatArray,
    bb_indices: IntArray,
    dim: int,
):
    Rmin = FLOAT_MAX
    Lmax = FLOAT_MIN
    for i in range(index, index + size):
        data_index = bb_indices[i]
        value = bb_coords[data_index, 2 * dim]
        if value < Rmin:
            Rmin = value
        value = bb_coords[data_index, 2 * dim + 1]
        if value > Lmax:
            Lmax = value
    return Rmin, Lmax


@nb.njit(inline="never", cache=True)
def split_plane(
    buckets: List[Bucket],
    root: np.void,
    range_Lmax: float,
    range_Rmin: float,
    bucket_length: float,
):
    plane_min_cost = FLOAT_MAX
    plane = INT_MAX
    bbs_in_left = 0
    bbs_in_right = 0

    # if we split here, lmax is from bucket 0, and rmin is from bucket 1 after
    # computing those, we can compute the cost to split here, and if this is the
    # minimum, we split here.
    for i in range(1, len(buckets)):
        current_bucket = buckets[i - 1]
        next_bucket = buckets[i]
        bbs_in_left += current_bucket.size
        bbs_in_right = root.size - bbs_in_left
        left_volume = (current_bucket.Lmax - range_Rmin) / bucket_length
        right_volume = (range_Lmax - next_bucket.Rmin) / bucket_length
        plane_cost = left_volume * bbs_in_left + right_volume * bbs_in_right
        if plane_cost < plane_min_cost:
            plane_min_cost = plane_cost
            plane = i

    Lmax = FLOAT_MIN
    Rmin = FLOAT_MAX
    for i in range(plane):
        bLmax = buckets[i].Lmax
        if bLmax > Lmax:
            Lmax = bLmax
    for i in range(plane, len(buckets)):
        bRmin = buckets[i].Rmin
        if bRmin < Rmin:
            Rmin = bRmin

    return plane, Lmax, Rmin


@nb.njit(cache=True)
def pessimistic_n_nodes(n_polys: int):
    """
    In the worst case, *all* branches end in a leaf with a single cell. Rather
    unlikely in the case of non-trivial grids, but we need a guess to
    pre-allocate -- overestimation is at maximum two times in case of
    cells_per_leaf == 2.
    """
    n_nodes = n_polys
    nodes = int(np.ceil(n_polys / 2))
    while nodes > 1:
        n_nodes += nodes
        nodes = int(np.ceil(nodes / 2))
    # Return, add root.
    return n_nodes + 1


@nb.njit(inline="always")
def push_both(root_stack, dim_stack, root, dim, size):
    size_root = push(root_stack, root, size)
    _ = push(dim_stack, dim, size)
    return size_root


@nb.njit(inline="always")
def pop_both(root_stack, dim_stack, size):
    root, size_root = pop(root_stack, size)
    dim, _ = pop(dim_stack, size)
    return root, dim, size_root


@nb.njit(cache=True)
def build(
    nodes: NodeArray,
    node_index: int,
    bb_indices: IntArray,
    bb_coords: FloatArray,
    n_buckets: int,
    cells_per_leaf: int,
):
    # Cannot compile ahead of time with Numba and recursion
    # Just use a stack based approach instead
    root_stack = allocate_stack()
    dim_stack = allocate_stack()
    root_stack[0] = 0
    dim_stack[0] = 0
    size = 1

    while size > 0:
        root_index, dim, size = pop_both(root_stack, dim_stack, size)

        dim_flag = dim
        if dim < 0:
            dim += 2

        # Fetch this root node
        root = Node(
            nodes[root_index]["child"],
            nodes[root_index]["Lmax"],
            nodes[root_index]["Rmin"],
            nodes[root_index]["ptr"],
            nodes[root_index]["size"],
            nodes[root_index]["dim"],
        )

        # Is it a leaf? if so, we're done, otherwise split.
        if root.size <= cells_per_leaf:
            continue

        # Find bounding range of node's entire dataset in dimension 0 (x-axis).
        range_Rmin, range_Lmax = get_bounds(
            root.ptr,
            root.size,
            bb_coords,
            bb_indices,
            dim,
        )
        bucket_length = (range_Lmax - range_Rmin) / n_buckets

        # Create buckets
        buckets = []
        # Specify ranges on the buckets
        for i in range(n_buckets):
            buckets.append(
                Bucket(
                    (i + 1) * bucket_length + range_Rmin,  # Max
                    i * bucket_length + range_Rmin,  # Min
                    -1.0,  # Rmin
                    -1.0,  # Lmax
                    -1,  # index
                    0,  # size
                )
            )
        # NOTA BENE: do not change the default size (0) given to the bucket here
        # it is used to detect empty buckets later on.

        # Now that the buckets are setup, sort them
        sort_bbox_indices(bb_indices, bb_coords, buckets, root, dim)

        # Determine Lmax and Rmin for each bucket
        for i in range(n_buckets):
            Rmin, Lmax = get_bounds(
                buckets[i].index, buckets[i].size, bb_coords, bb_indices, dim
            )
            b = buckets[i]
            buckets[i] = Bucket(b.Max, b.Min, Rmin, Lmax, b.index, b.size)

        # Special case: 2 bounding boxes share the same centroid, but boxes_per_leaf
        # is 1. This will break most of the usual bucketing code. Unless the grid has
        # overlapping triangles (which it shouldn't!). This is the only case to deal
        # with
        if (cells_per_leaf == 1) and (root.size == 2):
            nodes[root_index]["Lmax"] = range_Lmax
            nodes[root_index]["Rmin"] = range_Rmin
            left_child = create_node(root.ptr, 1, not dim)
            right_child = create_node(root.ptr + 1, 1, not dim)
            nodes[root_index]["child"] = node_index
            node_index = push_node(nodes, left_child, node_index)
            node_index = push_node(nodes, right_child, node_index)
            continue

        while buckets[0].size == 0:
            b = buckets[1]
            buckets[1] = Bucket(b.Max, buckets[0].Min, b.Rmin, b.Lmax, b.index, b.size)
            buckets.pop(0)

        i = 1
        while i < len(buckets):
            next_bucket = buckets[i]
            # if a empty bucket is encountered, merge it with the previous one and
            # continue as normal. As long as the ranges of the merged buckets are
            # still proper, calculating cost for empty buckets can be avoided, and
            # the split will still happen in the right place
            if next_bucket.size == 0:
                b = buckets[i - 1]
                buckets[i - 1] = Bucket(
                    next_bucket.Max, b.Min, b.Rmin, b.Lmax, b.index, b.size
                )
                buckets.pop(i)
            else:
                i += 1

        # Check if all the cells are in one bucket. If so, restart and switch
        # dimension.
        needs_continue = False
        for bucket in buckets:
            if bucket.size == root.size:
                needs_continue = True
                if dim_flag >= 0:
                    dim_flag = (not dim) - 2
                    nodes[root_index]["dim"] = not root.dim
                    size = push_both(root_stack, dim_stack, root_index, dim_flag, size)
                else:  # Already split once, convert to leaf.
                    nodes[root_index]["Lmax"] = -1
                    nodes[root_index]["Rmin"] = -1
                break
        if needs_continue:
            continue

        # plane is the separation line to split on:
        # 0 [bucket0] 1 [bucket1] 2 [bucket2] 3 [bucket3]
        plane, Lmax, Rmin = split_plane(
            buckets, root, range_Lmax, range_Rmin, bucket_length
        )
        right_index = buckets[plane].index
        right_size = root.ptr + root.size - right_index
        left_index = root.ptr
        left_size = root.size - right_size
        nodes[root_index]["Lmax"] = Lmax
        nodes[root_index]["Rmin"] = Rmin
        left_child = create_node(left_index, left_size, not dim)
        right_child = create_node(right_index, right_size, not dim)
        nodes[root_index]["child"] = node_index
        child_ind = node_index
        node_index = push_node(nodes, left_child, node_index)
        node_index = push_node(nodes, right_child, node_index)

        size = push_both(root_stack, dim_stack, child_ind + 1, right_child.dim, size)
        size = push_both(root_stack, dim_stack, child_ind, left_child.dim, size)

    return node_index


@nb.njit(cache=True)
def initialize(
    vertices: FloatArray, faces: IntArray, n_buckets: int = 4, cells_per_leaf: int = 2
) -> Tuple[NodeArray, IntArray]:
    # Prepare bounding boxes for tree building.
    bb_coords = build_bboxes(faces, vertices)
    bb_indices = np.arange(len(faces), dtype=IntDType)

    # Pre-allocate the space for the tree.
    n_polys, _ = faces.shape
    n_nodes = pessimistic_n_nodes(n_polys)
    nodes = np.empty(n_nodes, dtype=NodeDType)

    # Insert first node
    node = create_node(0, bb_indices.size, False)
    node_index = push_node(nodes, node, 0)

    node_index = build(
        nodes,
        node_index,
        bb_indices,
        bb_coords,
        n_buckets,
        cells_per_leaf,
    )

    # Remove the unused part in nodes.
    return nodes[:node_index], bb_indices, bb_coords



import matplotlib.pyplot as plt
import numpy as np
import pytest

from numba_celltree import demo


def test_close_polygons():
    faces = np.array(
        [
            [0, 1, 2, -1, -1],
            [0, 1, 2, 3, -1],
            [0, 1, 2, 3, 4],
        ]
    )
    closed = demo.close_polygons(faces, -1)
    expected = np.array(
        [
            [0, 1, 2, 0, 0, 0],
            [0, 1, 2, 3, 0, 0],
            [0, 1, 2, 3, 4, 0],
        ]
    )
    assert np.array_equal(closed, expected)


def test_edges():
    faces = np.array(
        [
            [0, 1, 2, -1],
            [1, 3, 4, 2],
        ]
    )
    actual = demo.edges(faces, -1)
    expected = np.array(
        [
            [0, 1],
            [0, 2],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
        ]
    )
    assert np.array_equal(actual, expected)


def test_plot_edges():
    _, ax = plt.subplots()
    node_x = np.array([0.0, 1.0, 1.0, 2.0, 2.0])
    node_y = np.array([0.0, 0.0, 1.0, 0.0, 1.0])
    edges = np.array(
        [
            [0, 1],
            [0, 2],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
        ]
    )
    demo.plot_edges(node_x, node_y, edges, ax)


def test_plot_boxes():
    boxes = np.array(
        [
            [0.0, 1.0, 0.0, 1.0],
            [1.0, 2.0, 1.0, 2.0],
        ]
    )
    _, ax = plt.subplots()
    demo.plot_boxes(boxes, ax, annotate=True)

    boxes = np.array(
        [
            [0.0, 1.0, 0.0],
            [1.0, 2.0, 1.0],
        ]
    )
    _, ax = plt.subplots()
    with pytest.raises(ValueError):
        demo.plot_boxes(boxes, ax)


def test_generate_disk():
    with pytest.raises(ValueError, match="partitions should be >= 3"):
        demo.generate_disk(2, 2)

    nodes, faces = demo.generate_disk(4, 1)
    assert nodes.shape == (5, 2)
    assert faces.shape == (4, 3)
    _, faces = demo.generate_disk(4, 2)
    assert faces.shape == (16, 3)



"""
This is a straightforward implementation of barycentric interpolation for
triangles.

For barycentric interpolation on N-sided convex polygons, see
barycentric_wachspress.py.
"""

import numba as nb
import numpy as np

from numba_celltree.constants import (
    PARALLEL,
    FloatArray,
    FloatDType,
    IntArray,
    Triangle,
)
from numba_celltree.geometry_utils import (
    Point,
    as_point,
    as_triangle,
    cross_product,
    to_vector,
)


@nb.njit(inline="always")
def compute_weights(triangle: Triangle, p: Point, weights: FloatArray):
    ab = to_vector(triangle.a, triangle.b)
    ac = to_vector(triangle.a, triangle.c)
    ap = to_vector(triangle.a, p)
    Aa = abs(cross_product(ab, ap))
    Ac = abs(cross_product(ac, ap))
    A = abs(cross_product(ab, ac))
    inv_denom = 1.0 / A
    w = inv_denom * Aa
    v = inv_denom * Ac
    u = 1.0 - v - w
    weights[0] = u
    weights[1] = v
    weights[2] = w
    return


@nb.njit(parallel=PARALLEL, cache=True)
def barycentric_triangle_weights(
    points: FloatArray,
    face_indices: IntArray,
    faces: IntArray,
    vertices: FloatArray,
) -> FloatArray:
    n_points = len(points)
    weights = np.zeros((n_points, 3), dtype=FloatDType)
    for i in nb.prange(n_points):
        face_index = face_indices[i]
        if face_index == -1:
            continue
        face = faces[face_index]
        triangle = as_triangle(vertices, face)
        point = as_point(points[i])
        compute_weights(triangle, point, weights[i])
    return weights



import numba as nb
import numpy as np

from numba_celltree import utils as ut
from numba_celltree.constants import MAX_N_VERTEX, MAX_TREE_DEPTH


def test_pop():
    size = 3
    stack = np.arange(10, 10 + size)
    v, size = ut.pop(stack, size)
    assert v == 12
    assert size == 2
    v, size = ut.pop(stack, size)
    assert v == 11
    assert size == 1
    v, size = ut.pop(stack, size)
    assert v == 10
    assert size == 0


def test_push():
    size = 0
    stack = np.empty(3)
    size = ut.push(stack, 0, size)
    size = ut.push(stack, 1, size)
    size = ut.push(stack, 2, size)
    assert size == 3
    assert np.array_equal(stack, [0, 1, 2])


def test_copy():
    dst = np.zeros(5)
    src = np.arange(5)
    ut.copy(src, dst, 3)
    assert np.array_equal(dst, [0, 1, 2, 0, 0])


# These array is not returned properly to dynamic python. This is OK: these
# arrays are exclusively for internal use to temporarily store values.
@nb.njit
def do_allocate_stack():
    stack = ut.allocate_stack()
    return (stack.size == MAX_TREE_DEPTH) and (stack[:5].size == 5)


def test_allocate_stack():
    assert do_allocate_stack()


@nb.njit
def do_allocate_polygon():
    poly = ut.allocate_polygon()
    return poly.shape == (MAX_N_VERTEX, 2) and (poly[:5].size == 10)


def test_allocate_polygon():
    assert do_allocate_polygon()


@nb.njit
def do_allocate_clipper():
    clipper = ut.allocate_clip_polygon()
    return clipper.shape == (MAX_N_VERTEX * 2, 2) and (clipper[:5].size == 10)


def test_allocate_clipper():
    assert do_allocate_clipper()



import os

import numba as nb
import numpy as np

from numba_celltree import geometry_utils as gu
from numba_celltree.constants import Box, Point, Triangle, Vector


def test_to_vector():
    a = Point(0.0, 0.0)
    b = Point(1.0, 2.0)
    actual = gu.to_vector(a, b)
    assert isinstance(actual, Vector)
    assert actual.x == 1.0
    assert actual.y == 2.0


def test_as_point():
    a = np.array([0.0, 1.0])
    actual = gu.as_point(a)
    assert isinstance(actual, Point)
    assert actual.x == 0.0
    assert actual.y == 1.0


def test_as_box():
    a = np.array([0.0, 1.0, 2.0, 3.0])
    actual = gu.as_box(a)
    assert isinstance(actual, Box)
    assert actual.xmin == 0.0
    assert actual.xmax == 1.0
    assert actual.ymin == 2.0
    assert actual.ymax == 3.0


def test_as_triangle():
    vertices = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    face = np.array([2, 0, 1])
    actual = gu.as_triangle(vertices, face)
    assert isinstance(actual, Triangle)
    assert actual.a == Point(1.0, 1.0)
    assert actual.b == Point(0.0, 0.0)
    assert actual.c == Point(1.0, 0.0)


def test_to_point():
    a = Point(0.0, 0.0)
    b = Point(1.0, 2.0)
    V = gu.to_vector(a, b)
    t = 0.0
    actual = gu.to_point(t, a, V)
    assert np.allclose(actual, a)

    t = 1.0
    actual = gu.to_point(t, a, V)
    assert np.allclose(actual, b)

    t = 0.5
    actual = gu.to_point(t, a, V)
    assert np.allclose(actual, Point(0.5, 1.0))


def test_cross_product():
    u = Vector(1.0, 2.0)
    v = Vector(3.0, 4.0)
    assert np.allclose(gu.cross_product(u, v), u.x * v.y - u.y * v.x)


def test_dot_product():
    u = Vector(1.0, 2.0)
    v = Vector(3.0, 4.0)
    assert np.allclose(gu.dot_product(u, v), np.dot(u, v))


def test_polygon_length():
    face = np.array([0, 1, 2])
    assert gu.polygon_length(face) == 3
    assert gu.polygon_length(face) == 3
    face = np.array([0, 1, 2, -1, -1])
    assert gu.polygon_length(face) == 3
    face = np.array([0, 1, 2, 3, -1])
    assert gu.polygon_length(face) == 4


def test_polygon_area():
    # square
    p = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    assert np.allclose(gu.polygon_area(p), 1.0)
    # triangle
    p = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    assert np.allclose(gu.polygon_area(p), 0.5)
    # pentagon, counter-clockwise
    p = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.5, 2.0],
            [0.0, 1.0],
        ]
    )
    assert np.allclose(gu.polygon_area(p), 1.5)
    # clockwise
    assert np.allclose(gu.polygon_area(p[::-1]), 1.5)


def test_point_in_polygon():
    poly = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    assert gu.point_in_polygon(Point(0.5, 0.25), poly)
    assert not gu.point_in_polygon(Point(1.5, 0.25), poly)

    assert gu.point_in_polygon(Point(0.0, 0.0), poly)
    assert gu.point_in_polygon(Point(0.0, 0.0), poly[::-1])
    assert gu.point_in_polygon(Point(0.5, 0.5), poly)
    assert gu.point_in_polygon(Point(0.5, 0.5), poly[::-1])
    assert not gu.point_in_polygon(Point(1.0, 1.0), poly)
    assert not gu.point_in_polygon(Point(1.0, 1.0), poly[::-1])


def test_boxes_intersect():
    # Identity
    a = Box(0.0, 1.0, 0.0, 1.0)
    b = a
    assert gu.boxes_intersect(a, b)
    assert gu.boxes_intersect(b, a)
    # Overlap
    b = Box(0.5, 1.5, 0.0, 1.0)
    assert gu.boxes_intersect(a, b)
    assert gu.boxes_intersect(b, a)
    # No overlap
    b = Box(1.5, 2.5, 0.5, 1.0)
    assert not gu.boxes_intersect(a, b)
    assert not gu.boxes_intersect(b, a)
    # Different identity
    b = a
    assert gu.boxes_intersect(a, b)
    assert gu.boxes_intersect(b, a)
    # Inside
    a = Box(0.0, 1.0, 0.0, 1.0)
    b = Box(0.25, 0.75, 0.25, 0.75)
    assert gu.boxes_intersect(a, b)
    assert gu.boxes_intersect(b, a)


def test_box_contained():
    a = Box(0.0, 1.0, 0.0, 1.0)
    b = Box(0.25, 0.75, 0.25, 0.75)
    assert gu.box_contained(a, a)
    assert gu.box_contained(b, a)
    assert not gu.box_contained(a, b)


def test_bounding_box():
    face = np.array([0, 1, 2])
    vertices = np.array(
        [
            [0.0, 1.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    assert gu.bounding_box(face, vertices) == (0.0, 1.0, 0.0, 1.0)
    face = np.array([0, 1, 2, -1, -1])
    assert gu.bounding_box(face, vertices) == (0.0, 1.0, 0.0, 1.0)


def test_build_bboxes():
    faces = np.array(
        [
            [0, 1, 2, -1],
            [0, 1, 2, 3],
        ]
    )
    vertices = np.array(
        [
            [0.0, 5.0],
            [5.0, 0.0],
            [5.0, 5.0],
            [0.0, 5.0],
        ]
    )
    expected = np.array(
        [
            [0.0, 5.0, 0.0, 5.0],
            [0.0, 5.0, 0.0, 5.0],
        ]
    )
    actual = gu.build_bboxes(faces, vertices)
    assert np.array_equal(actual, expected)


def test_copy_vertices():
    """
    This has to be tested inside of numba jitted function, because the vertices
    are copied to a stack allocated array. This array is not returned properly
    to dynamic python. This is OK: these arrays are exclusively for internal
    use to temporarily store values.
    """
    if os.environ.get("NUMBA_DISABLE_JIT", "0") == "0":

        @nb.njit()
        def test():
            face = np.array([0, 1, 2, -1, -1])
            vertices = np.array(
                [
                    [0.0, 1.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                ]
            )
            expected = vertices.copy()
            actual = gu.copy_vertices(vertices, face)
            result = True
            for i in range(3):
                result = result and actual[i, 0] == expected[i, 0]
                result = result and actual[i, 1] == expected[i, 1]
            return result

        assert test()

    else:
        face = np.array([0, 1, 2, -1, -1])
        vertices = np.array(
            [
                [0.0, 1.0],
                [1.0, 0.0],
                [1.0, 1.0],
            ]
        )
        expected = vertices.copy()
        actual = gu.copy_vertices(vertices, face)
        assert np.array_equal(actual, expected)
        assert len(actual) == 3


def test_copy_vertices_into():
    out = np.empty((10, 2))
    face = np.array([0, 1, 2, -1, -1])
    vertices = np.array(
        [
            [0.0, 1.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    expected = vertices.copy()
    actual = gu.copy_vertices_into(vertices, face, out)
    assert np.array_equal(actual, expected)
    assert len(actual) == 3


def test_point_inside_box():
    box = Box(0.0, 1.0, 0.0, 1.0)
    a = Point(0.5, 0.5)
    assert gu.point_inside_box(a, box)
    a = Point(-0.5, 0.5)
    assert not gu.point_inside_box(a, box)
    a = Point(0.5, -0.5)
    assert not gu.point_inside_box(a, box)


def test_flip():
    face0 = np.array([0, 1, 2, -1, -1])
    face1 = np.array([0, 1, 2, 3, -1])
    face2 = np.array([0, 1, 2, 3, 4])
    gu.flip(face0, 3)
    gu.flip(face1, 4)
    gu.flip(face2, 5)
    assert np.array_equal(face0, [2, 1, 0, -1, -1])
    assert np.array_equal(face1, [3, 2, 1, 0, -1])
    assert np.array_equal(face2, [4, 3, 2, 1, 0])


def test_counter_clockwise():
    vertices = np.array(
        [
            [0.0, 0.0],
            [0.5, 0.0],  # hanging node
            [1.0, 0.0],
            [1.0, 0.5],  # hanging node
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    ccw_faces = np.array(
        [
            [0, 2, 4, 5, -1, -1],
            [0, 1, 2, 3, 4, 5],
        ]
    )
    cw_faces = np.array(
        [
            [5, 4, 2, 0, -1, -1],
            [5, 4, 3, 2, 1, 0],
        ]
    )
    expected = ccw_faces.copy()
    # already counter clockwise should not be mutated
    gu.counter_clockwise(vertices, ccw_faces)
    assert np.array_equal(expected, ccw_faces)
    # clockwise should be mutated
    gu.counter_clockwise(vertices, cw_faces)
    assert np.array_equal(expected, cw_faces)



"""
Sutherland-Hodgman clipping
---------------------------
Vertices (always lower case, single letter):
Clipping polygon with vertices r, s, ...
Subject polgyon with vertices a, b, ...
Vectors (always upper case, single letter):

* U: r -> s
* N: norm, orthogonal to u
* V: a -> b
* W: a -> r

   s ----- ...
   |
   |  b ----- ...
   | /
   |/
   x
  /|
 / |
a--+-- ...
   |
   r ----- ...

Floating point rounding should not be an issue, since we're only looking at
finding the area of overlap of two convex polygons.
In case of intersection failure, we can ignore it when going out -> in. It will
occur when the outgoing point is very close the clipping edge. In that case the
intersection point ~= vertex b, and we can safely skip the intersection.
When going in -> out, b might be located on the edge. If intersection fails,
again the intersection point ~= vertex b. We treat b as if it is just on the
inside and append it. For consistency, we set b_inside to True, as it will be
used as a_inside in the next iteration.
"""

from typing import Sequence, Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import PARALLEL, FloatArray, FloatDType, IntArray
from numba_celltree.geometry_utils import (
    Point,
    Vector,
    as_box,
    as_point,
    copy_box_vertices,
    copy_vertices,
    dot_product,
    polygon_area,
)
from numba_celltree.utils import allocate_clip_polygon, copy


@nb.njit(inline="always")
def inside(p: Point, r: Point, U: Vector):
    # U: a -> b direction vector
    # p is point r or s
    return U.x * (p.y - r.y) > U.y * (p.x - r.x)


@nb.njit(inline="always")
def intersection(a: Point, V: Vector, r: Point, N: Vector) -> Tuple[bool, Point]:
    # Find the intersection with an (infinite) clipping plane
    W = Vector(r.x - a.x, r.y - a.y)
    nw = dot_product(N, W)
    nv = dot_product(N, V)
    if nv != 0:
        t = nw / nv
        return True, Point(a.x + t * V.x, a.y + t * V.y)
    else:
        # parallel lines
        return False, Point(np.nan, np.nan)


@nb.njit(inline="always")
def push_point(polygon: FloatArray, size: int, p: Point) -> int:
    polygon[size][0] = p.x
    polygon[size][1] = p.y
    return size + 1


@nb.njit(inline="always")
def polygon_polygon_clip_area(polygon: Sequence, clipper: Sequence) -> float:
    n_output = len(polygon)
    n_clip = len(clipper)
    subject = allocate_clip_polygon()
    output = allocate_clip_polygon()

    # Copy polygon into output
    copy(polygon, output, n_output)

    # Grab last point
    r = as_point(clipper[n_clip - 1])
    for i in range(n_clip):
        s = as_point(clipper[i])

        U = Vector(s.x - r.x, s.y - r.y)
        if U.x == 0 and U.y == 0:
            continue
        N = Vector(-U.y, U.x)

        # Copy output into subject
        length = n_output
        copy(output, subject, length)
        # Reset
        n_output = 0
        # Grab last point
        a = as_point(subject[length - 1])
        a_inside = inside(a, r, U)
        for j in range(length):
            b = as_point(subject[j])

            V = Vector(b.x - a.x, b.y - a.y)
            if V.x == 0 and V.y == 0:
                continue

            b_inside = inside(b, r, U)
            if b_inside:
                if not a_inside:  # out, or on the edge
                    succes, point = intersection(a, V, r, N)
                    if succes:
                        n_output = push_point(output, n_output, point)
                n_output = push_point(output, n_output, b)
            elif a_inside:
                succes, point = intersection(a, V, r, N)
                if succes:
                    n_output = push_point(output, n_output, point)
                else:  # Floating point failure
                    # TODO: haven't come up with a test case yet to succesfully
                    # trigger this ...
                    b_inside = True  # flip it for consistency, will be set as a
                    n_output = push_point(output, n_output, b)  # push b instead

            # Advance to next polygon edge
            a = b
            a_inside = b_inside

        # Exit early in case not enough vertices are left.
        if n_output < 3:
            return 0.0

        # Advance to next clipping edge
        r = s

    area = polygon_area(output[:n_output])
    return area


@nb.njit(parallel=PARALLEL, cache=True)
def area_of_intersection(
    vertices_a: FloatArray,
    vertices_b: FloatArray,
    faces_a: IntArray,
    faces_b: IntArray,
    indices_a: IntArray,
    indices_b: IntArray,
) -> FloatArray:
    n_intersection = indices_a.size
    area = np.empty(n_intersection, dtype=FloatDType)
    for i in nb.prange(n_intersection):
        face_a = faces_a[indices_a[i]]
        face_b = faces_b[indices_b[i]]
        a = copy_vertices(vertices_a, face_a)
        b = copy_vertices(vertices_b, face_b)
        area[i] = polygon_polygon_clip_area(a, b)
    return area


@nb.njit(parallel=PARALLEL, cache=True)
def box_area_of_intersection(
    bbox_coords: FloatArray,
    vertices: FloatArray,
    faces: IntArray,
    indices_bbox: IntArray,
    indices_face: IntArray,
) -> FloatArray:
    n_intersection = indices_bbox.size
    area = np.empty(n_intersection, dtype=FloatDType)
    for i in nb.prange(n_intersection):
        box = as_box(bbox_coords[indices_bbox[i]])
        face = faces[indices_face[i]]
        a = copy_box_vertices(box)
        b = copy_vertices(vertices, face)
        area[i] = polygon_polygon_clip_area(a, b)
    return area



from typing import Tuple

import matplotlib.tri
import numpy as np
from matplotlib import patches
from matplotlib.collections import LineCollection

from numba_celltree.constants import IntArray, IntDType


def close_polygons(face_node_connectivity: IntArray, fill_value: int) -> IntArray:
    # Wrap around and create closed polygon: put the first node at the end of the row
    # In case of fill values, replace all fill values
    n, m = face_node_connectivity.shape
    closed = np.full((n, m + 1), fill_value, dtype=IntDType)
    closed[:, :-1] = face_node_connectivity
    first_node = face_node_connectivity[:, 0]
    # Identify fill value, and replace by first node also
    isfill = closed == fill_value
    closed.ravel()[isfill.ravel()] = np.repeat(first_node, isfill.sum(axis=1))
    return closed


def edges(
    face_node_connectivity: IntArray, fill_value: int
) -> Tuple[IntArray, IntArray]:
    face_node_connectivity = np.atleast_2d(face_node_connectivity)
    n, m = face_node_connectivity.shape
    # Close the polygons: [0 1 2 3] -> [0 1 2 3 0]
    closed = close_polygons(face_node_connectivity, fill_value)
    # Allocate array for edge_node_connectivity: includes duplicate edges
    edge_node_connectivity = np.empty((n * m, 2), dtype=IntDType)
    edge_node_connectivity[:, 0] = closed[:, :-1].ravel()
    edge_node_connectivity[:, 1] = closed[:, 1:].ravel()
    # Cleanup: delete invalid edges (same node to same node)
    # this is a result of closing the polygons
    edge_node_connectivity = edge_node_connectivity[
        edge_node_connectivity[:, 0] != edge_node_connectivity[:, 1]
    ]
    # Now find the unique rows == unique edges
    edge_node_connectivity.sort(axis=1)
    edge_node_connectivity = np.unique(edge_node_connectivity, axis=0)
    return edge_node_connectivity


def plot_edges(node_x, node_y, edge_nodes, ax, *args, **kwargs):
    """
    Plot the edges at a given axes.
    `args` and `kwargs` will be used as parameters of the `plot` method.

    Parameters
    ----------
    node_x (ndarray): A 1D double array describing the x-coordinates of the nodes.
    node_y (ndarray): A 1D double array describing the y-coordinates of the nodes.
    edge_nodes (ndarray, optional): A 2D integer array describing the nodes composing each mesh 2d edge.
    ax (matplotlib.axes.Axes): The axes where to plot the edges
    """
    n_edge = len(edge_nodes)
    edge_coords = np.empty((n_edge, 2, 2), dtype=np.float64)
    node_0 = edge_nodes[:, 0]
    node_1 = edge_nodes[:, 1]
    edge_coords[:, 0, 0] = node_x[node_0]
    edge_coords[:, 0, 1] = node_y[node_0]
    edge_coords[:, 1, 0] = node_x[node_1]
    edge_coords[:, 1, 1] = node_y[node_1]
    line_segments = LineCollection(edge_coords, *args, **kwargs)
    ax.add_collection(line_segments)
    ax.set_aspect(1.0)
    ax.autoscale(enable=True)
    return


def plot_boxes(box_coords, ax, annotate=False, *args, **kwargs):
    box_coords = np.atleast_2d(box_coords)
    nbox, ncoord = box_coords.shape
    if ncoord != 4:
        raise ValueError(f"four values describe a box, got instead {ncoord}")
    for i in range(nbox):
        xmin, xmax, ymin, ymax = box_coords[i]
        dx = xmax - xmin
        dy = ymax - ymin
        rect = patches.Rectangle((xmin, ymin), dx, dy, fill=False, *args, **kwargs)
        ax.add_patch(rect)
        if annotate:
            ax.annotate(i, (xmin + 0.5 * dx, ymin + 0.5 * dy))
    ax.set_aspect(1.0)
    ax.autoscale(enable=True)
    return


def generate_disk(partitions: int, depth: int):
    """
    Generate a triangular mesh for the unit circle.

    Parameters
    ----------
    partitions: int
        Number of triangles around the origin.
    depth: int
        Number of "layers" of triangles around the origin.

    Returns
    -------
    vertices: np.ndarray of floats with shape ``(n_vertex, 2)``
    triangles: np.ndarray of integers with shape ``(n_triangle, 3)``
    """
    if partitions < 3:
        raise ValueError("partitions should be >= 3")

    N = depth + 1
    n_per_level = partitions * np.arange(N)
    n_per_level[0] = 1

    delta_angle = (2 * np.pi) / np.repeat(n_per_level, n_per_level)
    index = np.repeat(np.insert(n_per_level.cumsum()[:-1], 0, 0), n_per_level)
    angles = delta_angle.cumsum()
    angles = angles - angles[index] + 0.5 * np.pi
    radii = np.repeat(np.linspace(0.0, 1.0, N), n_per_level)

    x = np.cos(angles) * radii
    y = np.sin(angles) * radii
    triang = matplotlib.tri.Triangulation(x, y)
    return np.column_stack((x, y)), triang.triangles



import numpy as np
import pytest

from numba_celltree.algorithms import (
    cohen_sutherland_line_box_clip,
    cyrus_beck_line_polygon_clip,
    liang_barsky_line_box_clip,
)
from numba_celltree.constants import Box, Point


def ab(a, b, c):
    """Flip the result around to compare (a, b) with (b, a)"""
    return (a, c, b)


BOX = Box(0.0, 2.0, 0.0, 2.0)
POLY = np.array(
    [
        [0.0, 0.0],
        [2.0, 0.0],
        [2.0, 2.0],
        [0.0, 2.0],
    ]
)


@pytest.mark.parametrize(
    "line_clip, box",
    [
        (cohen_sutherland_line_box_clip, BOX),
        (liang_barsky_line_box_clip, BOX),
        (cyrus_beck_line_polygon_clip, POLY),
    ],
)
def test_line_box_clip(line_clip, box):
    a = Point(-1.0, 0.0)
    b = Point(2.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [1.0, 2.0])

    a = Point(0.0, -0.1)
    b = Point(0.0, -0.1)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects
    assert np.isnan(c).all()
    assert np.isnan(d).all()

    a = Point(-1.0, 1.0)
    b = Point(3.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [2.0, 1.0])

    a = Point(1.0, -3.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 2.0])

    b = Point(1.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 1.0])

    a = Point(1.0, 1.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 1.0])
    assert np.allclose(d, [1.0, 2.0])

    a = Point(-1.0, 3.0)
    b = Point(3.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects

    a = Point(-1.0, 1.0)
    b = Point(1.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [1.0, 1.0])

    # both inside
    a = Point(0.5, 0.5)
    b = Point(1.5, 1.5)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, a)
    assert np.allclose(d, b)

    # No intersection, left
    a = Point(-1.5, 0.0)
    b = Point(-0.5, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects

    # No intersection, right
    a = Point(2.5, 0.0)
    b = Point(3.5, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects


@pytest.mark.parametrize(
    "line_clip, box",
    [
        (cohen_sutherland_line_box_clip, BOX),
        (liang_barsky_line_box_clip, BOX),
        (cyrus_beck_line_polygon_clip, POLY),
    ],
)
def test_line_box_clip_degeneracy(line_clip, box):
    # Line through vertices
    a = Point(-1.0, -1.0)
    b = Point(3.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 0.0])
    assert np.allclose(d, [2.0, 2.0])
    assert line_clip(a, b, box)[0] == line_clip(b, a, box)[0]

    # Identity
    intersects, c, d = line_clip(a, a, box)
    assert not intersects

    # Line through lower edge
    a = Point(-1.0, 0.0)
    b = Point(3.0, 0.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Line through upper edge
    a = Point(-1.0, 2.0)
    b = Point(3.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Partial line lower edge
    a = Point(-1.0, 0.0)
    b = Point(1.0, 0.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Partial line upper edge
    a = Point(-1.0, 2.0)
    b = Point(1.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Within lower edge
    a = Point(0.5, 0.0)
    b = Point(1.5, 0.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Within upper edge
    a = Point(0.5, 2.0)
    b = Point(1.5, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to lower edge
    a = Point(0.0, 0.0)
    b = Point(2.0, 0.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to upper edge
    a = Point(0.0, 2.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to left edge
    a = Point(0.0, 0.0)
    b = Point(0.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to right edge
    a = Point(2.0, 0.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Within left edge
    a = Point(0.0, 0.5)
    b = Point(0.0, 1.5)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Within right edge
    a = Point(2.0, 0.5)
    b = Point(2.0, 1.5)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to left edge
    a = Point(0.0, 0.0)
    b = Point(0.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]

    # Identical to right edge
    a = Point(2.0, 0.0)
    b = Point(2.0, 2.0)
    assert line_clip(a, b, box)[0]
    assert line_clip(b, a, box)[0]



import os

import numba as nb
import numpy as np
from numba import types
from numba.core import cgutils
from numba.extending import intrinsic

from numba_celltree.constants import (
    MAX_N_VERTEX,
    MAX_TREE_DEPTH,
    NDIM,
    STACK_ALLOCATE_STATIC_ARRAYS,
    FloatDType,
    IntDType,
)


@nb.njit(inline="always")
def pop(array, size):
    return array[size - 1], size - 1


@nb.njit(inline="always")
def push(array, value, size):
    array[size] = value
    return size + 1


@nb.njit(inline="always")
def copy(src, dst, n) -> None:
    for i in range(n):
        dst[i] = src[i]


# Ensure these are constants for numba
POLYGON_SIZE = MAX_N_VERTEX * NDIM
CLIP_MAX_N_VERTEX = MAX_N_VERTEX * 2
CLIP_POLYGON_SIZE = 2 * POLYGON_SIZE


# Make sure everything still works when calling as non-compiled Python code.
# Note: these stack allocated arrays should only be used inside of numba
# compiled code. They should interact NEVER with dynamic Python code: there are
# no guarantees in that case, they may very well be filled with garbage.
if STACK_ALLOCATE_STATIC_ARRAYS and os.environ.get("NUMBA_DISABLE_JIT", "0") == "0":

    @intrinsic  # pragma: no cover
    def stack_empty(typingctx, size, dtype):
        def impl(context, builder, signature, args):
            ty = context.get_value_type(dtype.dtype)
            ptr = cgutils.alloca_once(builder, ty, size=args[0])
            return ptr

        sig = types.CPointer(dtype.dtype)(types.int64, dtype)
        return sig, impl

    @nb.njit(inline="always")  # pragma: no cover
    def allocate_stack():
        arr_ptr = stack_empty(MAX_TREE_DEPTH, IntDType)  # pylint: disable=no-value-for-parameter
        arr = nb.carray(arr_ptr, MAX_TREE_DEPTH, dtype=IntDType)
        return arr

    @nb.njit(inline="always")  # pragma: no cover
    def allocate_polygon():
        arr_ptr = stack_empty(  # pylint: disable=no-value-for-parameter
            POLYGON_SIZE, FloatDType
        )
        arr = nb.carray(arr_ptr, (MAX_N_VERTEX, NDIM), dtype=FloatDType)
        return arr

    @nb.njit(inline="always")  # pragma: no cover
    def allocate_clip_polygon():
        arr_ptr = stack_empty(  # pylint: disable=no-value-for-parameter
            CLIP_POLYGON_SIZE, FloatDType
        )
        arr = nb.carray(arr_ptr, (CLIP_MAX_N_VERTEX, NDIM), dtype=FloatDType)
        return arr

    @nb.njit(inline="always")  # pragma: no cover
    def allocate_box_polygon():
        arr_ptr = stack_empty(8, FloatDType)  # pylint: disable=no-value-for-parameter
        arr = nb.carray(arr_ptr, (4, 2), dtype=FloatDType)
        return arr

else:

    @nb.njit(inline="always")
    def allocate_stack():
        return np.empty(MAX_TREE_DEPTH, dtype=IntDType)

    @nb.njit(inline="always")
    def allocate_polygon():
        return np.empty((MAX_N_VERTEX, NDIM), dtype=FloatDType)

    @nb.njit(inline="always")
    def allocate_clip_polygon():
        return np.empty((CLIP_MAX_N_VERTEX, NDIM), dtype=FloatDType)

    @nb.njit(inline="always")
    def allocate_box_polygon():
        return np.empty((4, 2), dtype=FloatDType)



from typing import Tuple

import numpy as np

from numba_celltree.algorithms import (
    area_of_intersection,
    barycentric_triangle_weights,
    barycentric_wachspress_weights,
    box_area_of_intersection,
    polygons_intersect,
)
from numba_celltree.constants import (
    FILL_VALUE,
    MAX_N_VERTEX,
    BoolArray,
    CellTreeData,
    FloatArray,
    FloatDType,
    IntArray,
    IntDType,
)
from numba_celltree.creation import initialize
from numba_celltree.geometry_utils import build_bboxes, counter_clockwise
from numba_celltree.query import (
    collect_node_bounds,
    locate_boxes,
    locate_edges,
    locate_points,
    validate_node_bounds,
)


# Ensure all types are as as statically expected.
def cast_vertices(vertices: FloatArray, copy: bool = False) -> FloatArray:
    if isinstance(vertices, np.ndarray):
        vertices = vertices.astype(FloatDType, copy=copy)
    else:
        vertices = np.ascontiguousarray(vertices, dtype=FloatDType)
    if vertices.ndim != 2 or vertices.shape[1] != 2:
        raise ValueError("vertices must have shape (n_points, 2)")
    return vertices


def cast_faces(faces: IntArray, fill_value: int) -> IntArray:
    if isinstance(faces, np.ndarray):
        faces = faces.astype(IntDType, copy=True)
    else:
        faces = np.ascontiguousarray(faces, dtype=IntDType)
    if faces.ndim != 2:
        raise ValueError("faces must have shape (n_face, n_max_vert)")
    _, n_max_vert = faces.shape
    if n_max_vert > MAX_N_VERTEX:
        raise ValueError(
            f"faces contains up to {n_max_vert} vertices for a single face. "
            f"numba_celltree supports a maximum of {MAX_N_VERTEX} vertices. "
            f"Increase MAX_N_VERTEX in the source code, or alter the mesh."
        )
    if fill_value != FILL_VALUE:
        faces[faces == fill_value] = FILL_VALUE
    return faces


def cast_bboxes(bbox_coords: FloatArray) -> FloatArray:
    bbox_coords = np.ascontiguousarray(bbox_coords, dtype=FloatDType)
    if bbox_coords.ndim != 2 or bbox_coords.shape[1] != 4:
        raise ValueError("bbox_coords must have shape (n_box, 4)")
    return bbox_coords


def cast_edges(edges: FloatArray) -> FloatArray:
    edges = np.ascontiguousarray(edges, dtype=FloatDType)
    if edges.ndim != 3 or edges.shape[1] != 2 or edges.shape[2] != 2:
        raise ValueError("edges must have shape (n_edge, 2, 2)")
    return edges


def bbox_tree(bb_coords: FloatArray) -> FloatArray:
    xmin = bb_coords[:, 0].min()
    xmax = bb_coords[:, 1].max()
    ymin = bb_coords[:, 2].min()
    ymax = bb_coords[:, 3].max()
    return np.array([xmin, xmax, ymin, ymax], dtype=FloatDType)


class CellTree2d:
    """
    Construct a cell tree from 2D vertices and a faces indexing array.

    Parameters
    ----------
    vertices: ndarray of floats with shape ``(n_point, 2)``
        Corner coordinates (x, y) of the cells.
    faces: ndarray of integers with shape ``(n_face, n_max_vert)``
        Index identifying for every face the indices of its corner nodes.  If a
        face has less corner nodes than ``n_max_vert``, its last indices should
        be equal to ``fill_value``.
    n_buckets: int, optional, default: 4
        The number of "buckets" used in tree construction. Must be higher
        or equal to 2. Values over 8 provide diminishing returns.
    cells_per_leaf: int, optional, default: 2
        The number of cells in the leaf nodes of the cell tree. Can be set
        to only 1, but this doubles memory footprint for slightly faster
        lookup. Increase this to reduce memory usage at the cost of lookup
        performance.
    fill_value: int, optional, default: -1
        Fill value marking empty nodes in ``faces``.
    """

    def __init__(
        self,
        vertices: FloatArray,
        faces: IntArray,
        fill_value: int,
        n_buckets: int = 4,
        cells_per_leaf: int = 2,
    ):
        if n_buckets < 2:
            raise ValueError("n_buckets must be >= 2")
        if cells_per_leaf < 1:
            raise ValueError("cells_per_leaf must be >= 1")

        vertices = cast_vertices(vertices, copy=True)
        faces = cast_faces(faces, fill_value)
        counter_clockwise(vertices, faces)

        nodes, bb_indices, bb_coords = initialize(
            vertices, faces, n_buckets, cells_per_leaf
        )
        self.vertices = vertices
        self.faces = faces
        self.n_buckets = n_buckets
        self.cells_per_leaf = cells_per_leaf
        self.nodes = nodes
        self.bb_indices = bb_indices
        self.bb_coords = bb_coords
        self.bbox = bbox_tree(bb_coords)
        self.celltree_data = CellTreeData(
            self.faces,
            self.vertices,
            self.nodes,
            self.bb_indices,
            self.bb_coords,
            self.bbox,
            self.cells_per_leaf,
        )

    def locate_points(self, points: FloatArray) -> IntArray:
        """
        Find the index of a face that contains a point.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``

        Returns
        -------
        tree_face_indices: ndarray of integers with shape ``(n_point,)``
            For every point, the index of the face it falls in. Points not
            falling in any faces are marked with a value of ``-1``.
        """
        points = cast_vertices(points)
        return locate_points(points, self.celltree_data)

    def locate_boxes(self, bbox_coords: FloatArray) -> Tuple[IntArray, IntArray]:
        """
        Find the index of a face intersecting with a bounding box.

        Parameters
        ----------
        bbox_coords: ndarray of floats with shape ``(n_box, 4)``
            Every row containing ``(xmin, xmax, ymin, ymax)``.

        Returns
        -------
        bbox_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the bounding box.
        tree_face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the face.
        """
        bbox_coords = cast_bboxes(bbox_coords)
        return locate_boxes(bbox_coords, self.celltree_data)

    def intersect_boxes(
        self, bbox_coords: FloatArray
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Find the index of a box intersecting with a face, and the area
        of intersection.

        Parameters
        ----------
        bbox_coords: ndarray of floats with shape ``(n_box, 4)``
            Every row containing ``(xmin, xmax, ymin, ymax)``.

        Returns
        -------
        bbox_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the bounding box.
        tree_face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the tree faces.
        area: ndarray of floats with shape ``(n_found,)``
            Area of intersection between the two intersecting faces.
        """
        bbox_coords = cast_bboxes(bbox_coords)
        i, j = locate_boxes(bbox_coords, self.celltree_data)
        area = box_area_of_intersection(
            bbox_coords=bbox_coords,
            vertices=self.vertices,
            faces=self.faces,
            indices_bbox=i,
            indices_face=j,
        )
        # Separating axes declares polygons with shared edges as touching.
        # Make sure we only include actual intersections.
        actual = area > 0
        return i[actual], j[actual], area[actual]

    def locate_faces(
        self, vertices: FloatArray, faces: IntArray
    ) -> Tuple[IntArray, IntArray]:
        """
        Find the index of a face intersecting with another face.

        Only sharing an edge also counts as an intersection, due to the use of
        the separating axis theorem to define intersection. The area of the
        overlap is zero in such a case.

        Parameters
        ----------
        vertices: ndarray of floats with shape ``(n_point, 2)``
            Corner coordinates (x, y) of the cells.
        faces: ndarray of integers with shape ``(n_face, n_max_vert)``
            Index identifying for every face the indices of its corner nodes.
            If a face has less corner nodes than n_max_vert, its last indices
            should be equal to ``fill_value``.

        Returns
        -------
        face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the faces.
        tree_face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the tree faces.
        """
        counter_clockwise(vertices, faces)
        bbox_coords = build_bboxes(faces, vertices)
        shortlist_i, shortlist_j = locate_boxes(bbox_coords, self.celltree_data)
        intersects = polygons_intersect(
            vertices_a=vertices,
            vertices_b=self.vertices,
            faces_a=faces,
            faces_b=self.faces,
            indices_a=shortlist_i,
            indices_b=shortlist_j,
        )
        return shortlist_i[intersects], shortlist_j[intersects]

    def intersect_faces(
        self, vertices: FloatArray, faces: IntArray, fill_value: int
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Find the index of a face intersecting with another face, and the area
        of intersection.

        Parameters
        ----------
        vertices: ndarray of floats with shape ``(n_point, 2)``
            Corner coordinates (x, y) of the cells.
        faces: ndarray of integers with shape ``(n_face, n_max_vert)``
            Index identifying for every face the indices of its corner nodes.
            If a face has less corner nodes than n_max_vert, its last indices
            should be equal to ``fill_value``.
        fill_value: int, optional, default: -1
            Fill value marking empty nodes in ``faces``.

        Returns
        -------
        face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the faces.
        tree_face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the tree faces.
        area: ndarray of floats with shape ``(n_found,)``
            Area of intersection between the two intersecting faces.
        """
        vertices = cast_vertices(vertices)
        faces = cast_faces(faces, fill_value)
        i, j = self.locate_faces(vertices, faces)
        area = area_of_intersection(
            vertices_a=vertices,
            vertices_b=self.vertices,
            faces_a=faces,
            faces_b=self.faces,
            indices_a=i,
            indices_b=j,
        )
        # Separating axes declares polygons with shared edges as touching.
        # Make sure we only include actual intersections.
        actual = area > 0
        return i[actual], j[actual], area[actual]

    def intersect_edges(
        self, edge_coords: FloatArray
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Find the index of a face intersecting with an edge.

        Parameters
        ----------
        edge_coords: ndarray of floats with shape ``(n_edge, 2, 2)``
            Every row containing ``((x0, y0), (x1, y1))``.

        Returns
        -------
        edge_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the bounding box.
        tree_face_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the face.
        intersection_edges: ndarray of floats with shape ``(n_found, 2, 2)``
            The resulting intersected edges, every row containing:
            ``((x0, y0), (x1, y1))``.
            The length of each intersected edge can be computed with:
            ``np.linalg.norm(intersections[:, 1] - intersections[:, 0], axis=1)``.
        """
        edge_coords = cast_edges(edge_coords)
        return locate_edges(edge_coords, self.celltree_data)

    def compute_barycentric_weights(
        self,
        points: FloatArray,
    ) -> Tuple[IntArray, FloatArray]:
        """
        Compute barycentric weights for points located inside of the grid.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``

        Returns
        -------
        tree_face_indices: ndarray of integers with shape ``(n_point,)``
            For every point, the index of the face it falls in. Points not
            falling in any faces are marked with a value of ``-1``.
        barycentric_weights: ndarray of integers with shape ``(n_point, n_max_vert)``
            For every point, the barycentric weights of the vertices of the
            face in which the point is located. For points not falling in any
            faces, the weight of all vertices is 0.
        """
        face_indices = self.locate_points(points)
        n_max_vert = self.faces.shape[1]
        if n_max_vert > 3:
            f = barycentric_wachspress_weights
        else:
            f = barycentric_triangle_weights

        weights = f(
            points,
            face_indices,
            self.faces,
            self.vertices,
        )
        return face_indices, weights

    @property
    def node_bounds(self):
        """Return the bounds (xmin, xmax, ymin, ymax) for every node of the tree."""
        return collect_node_bounds(self.celltree_data)

    def validate_node_bounds(self) -> BoolArray:
        """
        Traverse the tree. Check whether all children are contained in the bounding
        box.

        For the leaf nodes, check whether the bounding boxes are contained.

        Returns
        -------
        node_validity: np.array of bool
            For each node, whether all children are fully contained by its
            bounds.
        """
        return validate_node_bounds(self.celltree_data, self.node_bounds)

    def to_dict_of_lists(self):
        """
        Convert the tree structure to a dict of lists.

        Such a dict can be ingested by e.g. NetworkX to produce visualize the
        tree structure.

        Returns
        -------
        dict_of_lists: Dict[Int, List[Int]]
            Contains for every node a list with its children.

        Examples
        --------
        >>> import networkx
        >>> from networkx.drawing.nx_pydot import graphviz_layout
        >>> d = celltree.to_dict_of_lists()
        >>> G = networkx.DiGraph(d)
        >>> positions = graphviz_layout(G, prog="dot")
        >>> networkx.draw(G, positions, with_labels=True)

        Note that computing the graphviz layout may be quite slow!
        """
        dict_of_lists = {}
        for parent_index, node in enumerate(self.celltree_data.nodes):
            left_child = node["child"]
            if left_child == -1:
                dict_of_lists[parent_index] = []
            else:
                right_child = left_child + 1
                dict_of_lists[parent_index] = [left_child, right_child]

        return dict_of_lists



import numba as nb
import numpy as np

from numba_celltree.algorithms import (
    cohen_sutherland_line_box_clip,
    cyrus_beck_line_polygon_clip,
)
from numba_celltree.constants import (
    PARALLEL,
    BoolArray,
    CellTreeData,
    FloatArray,
    FloatDType,
    IntArray,
    IntDType,
)
from numba_celltree.geometry_utils import (
    Box,
    Point,
    as_box,
    as_point,
    box_contained,
    boxes_intersect,
    copy_vertices_into,
    point_in_polygon,
    to_vector,
)
from numba_celltree.utils import allocate_polygon, allocate_stack, pop, push


# Inlining saves about 15% runtime
@nb.njit(inline="always")
def locate_point(point: Point, tree: CellTreeData):
    stack = allocate_stack()
    polygon_work_array = allocate_polygon()
    stack[0] = 0
    return_value = -1
    size = 1

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]

        # Check if it's a leaf
        if node["child"] == -1:
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                face = tree.faces[bbox_index]
                # Make sure polygons to test is contiguous (stack allocated) array
                # This saves about 40-50% runtime
                poly = copy_vertices_into(tree.vertices, face, polygon_work_array)
                if point_in_polygon(point, poly):
                    return bbox_index
            continue

        dim = 1 if node["dim"] else 0
        left = point[dim] <= node["Lmax"]
        right = point[dim] >= node["Rmin"]
        left_child = node["child"]
        right_child = left_child + 1

        if left and right:
            # This heuristic is worthwhile because a point will fall into a
            # single face -- if found, we can stop.
            if (node["Lmax"] - point[dim]) < (point[dim] - node["Rmin"]):
                size = push(stack, left_child, size)
                size = push(stack, right_child, size)
            else:
                size = push(stack, right_child, size)
                size = push(stack, left_child, size)
        elif left:
            size = push(stack, left_child, size)
        elif right:
            size = push(stack, right_child, size)

    return return_value


@nb.njit(parallel=PARALLEL, cache=True)
def locate_points(
    points: FloatArray,
    tree: CellTreeData,
):
    n_points = len(points)
    result = np.empty(n_points, dtype=IntDType)
    for i in nb.prange(n_points):  # pylint: disable=not-an-iterable
        point = as_point(points[i])
        result[i] = locate_point(point, tree)
    return result


@nb.njit(inline="always")
def locate_box(box: Box, tree: CellTreeData, indices: IntArray, store_indices: bool):
    tree_bbox = as_box(tree.bbox)
    if not boxes_intersect(box, tree_bbox):
        return 0
    stack = allocate_stack()
    stack[0] = 0
    size = 1
    count = 0

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]
        # Check if it's a leaf
        if node["child"] == -1:
            # Iterate over the bboxes in the leaf
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                # As a named tuple: saves about 15% runtime
                leaf_box = as_box(tree.bb_coords[bbox_index])
                if boxes_intersect(box, leaf_box):
                    if store_indices:
                        indices[count] = bbox_index
                    count += 1
        else:
            dim = 1 if node["dim"] else 0
            minimum = 2 * dim
            maximum = 2 * dim + 1
            left = box[minimum] <= node["Lmax"]
            right = box[maximum] >= node["Rmin"]
            left_child = node["child"]
            right_child = left_child + 1

            if left and right:
                size = push(stack, left_child, size)
                size = push(stack, right_child, size)
            elif left:
                size = push(stack, left_child, size)
            elif right:
                size = push(stack, right_child, size)

    return count


@nb.njit(parallel=PARALLEL, cache=True)
def locate_boxes(
    box_coords: FloatArray,
    tree: CellTreeData,
):
    # Numba does not support a concurrent list or bag like stucture:
    # https://github.com/numba/numba/issues/5878
    # (Standard lists are not thread safe.)
    # To support parallel execution, we're stuck with numpy arrays therefore.
    # Since we don't know the number of contained bounding boxes, we traverse
    # the tree twice: first to count, then allocate, then another time to
    # actually store the indices.
    # The cost of traversing twice is roughly a factor two. Since many
    # computers can parallellize over more than two threads, counting first --
    # which enables parallelization -- should still result in a net speed up.
    n_box = box_coords.shape[0]
    counts = np.empty(n_box + 1, dtype=IntDType)
    dummy = np.empty((0,), dtype=IntDType)
    counts[0] = 0
    # First run a count so we can allocate afterwards
    for i in nb.prange(n_box):  # pylint: disable=not-an-iterable
        box = as_box(box_coords[i])
        counts[i + 1] = locate_box(box, tree, dummy, False)

    # Run a cumulative sum
    total = 0
    for i in range(1, n_box + 1):
        total += counts[i]
        counts[i] = total

    # Now allocate appropriately
    ii = np.empty(total, dtype=IntDType)
    jj = np.empty(total, dtype=IntDType)
    for i in nb.prange(n_box):  # pylint: disable=not-an-iterable
        start = counts[i]
        end = counts[i + 1]
        ii[start:end] = i
        indices = jj[start:end]
        box = as_box(box_coords[i])
        locate_box(box, tree, indices, True)

    return ii, jj


# Inlining this function drives compilation time through the roof. It's
# probably also a rather bad idea, given its complexity: compared to looking
# for either boxes or points, checking is more much complicated by involving
# two intersection algorithms.
@nb.njit(inline="never")
def locate_edge(
    a: Point,
    b: Point,
    tree: CellTreeData,
    indices: IntArray,
    intersections: FloatArray,
    store_intersection: bool,
):
    # Check if the entire mesh intersects with the line segment at all
    tree_bbox = as_box(tree.bbox)
    tree_intersects, _, _ = cohen_sutherland_line_box_clip(a, b, tree_bbox)
    if not tree_intersects:
        return 0

    V = to_vector(a, b)
    stack = allocate_stack()
    polygon_work_array = allocate_polygon()
    stack[0] = 0
    size = 1
    count = 0

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]

        # Check if it's a leaf
        if node["child"] == -1:
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                box = as_box(tree.bb_coords[bbox_index])
                box_intersect, _, _ = cohen_sutherland_line_box_clip(a, b, box)
                if box_intersect:
                    polygon = copy_vertices_into(
                        tree.vertices, tree.faces[bbox_index], polygon_work_array
                    )
                    face_intersects, c, d = cyrus_beck_line_polygon_clip(a, b, polygon)
                    if face_intersects:
                        if store_intersection:
                            indices[count] = bbox_index
                            intersections[count, 0, 0] = c.x
                            intersections[count, 0, 1] = c.y
                            intersections[count, 1, 0] = d.x
                            intersections[count, 1, 1] = d.y
                        count += 1
            continue

        # Note, "x" is a placeholder for x, y here
        # Contrast with t, which is along vector
        node_dim = 1 if node["dim"] else 0
        dx = V[node_dim]
        if dx > 0.0:
            dx_left = node["Lmax"] - a[node_dim]
            dx_right = node["Rmin"] - b[node_dim]
        else:
            dx_left = node["Lmax"] - b[node_dim]
            dx_right = node["Rmin"] - a[node_dim]

        # Check how origin (a) and end (b) are located compared to box edges
        # (Lmax, Rmin). The box should be investigated if:
        # * the origin is left of Lmax (dx_left >= 0)
        # * the end is right of Rmin (dx_right <= 0)
        left = dx_left >= 0.0
        right = dx_right <= 0.0

        # Now find the intersection coordinates. These have to occur within in
        # the bounds of the vector. Note that if the line has no slope in this
        # dim (dx == 0), we cannot compute the intersection, and we have to
        # defer to the child nodes.
        if dx > 0.0:  # TODO: abs(dx) > EPISLON?
            if left:
                t_left = dx_left / dx
                left = t_left >= 0.0
            if right:
                t_right = dx_right / dx
                right = t_right <= 1.0
        elif dx < 0.0:
            if left:
                t_left = 1.0 - (dx_left / dx)
                left = t_left >= 0.0
            if right:
                t_right = 1.0 - (dx_right / dx)
                right = t_right <= 1.0
        # else dx == 0.0. In this case there's no info to extract from this
        # node. We'll fully defer to the children.
        left_child = node["child"]
        right_child = left_child + 1

        if left and right:
            size = push(stack, left_child, size)
            size = push(stack, right_child, size)
        elif left:
            size = push(stack, left_child, size)
        elif right:
            size = push(stack, right_child, size)

    return count


@nb.njit(parallel=PARALLEL, cache=True)
def locate_edges(
    edge_coords: FloatArray,
    tree: CellTreeData,
):
    # Numba does not support a concurrent list or bag like stucture:
    # https://github.com/numba/numba/issues/5878
    # (Standard lists are not thread safe.)
    # To support parallel execution, we're stuck with numpy arrays therefore.
    # Since we don't know the number of contained bounding boxes, we traverse
    # the tree twice: first to count, then allocate, then another time to
    # actually store the indices.
    # The cost of traversing twice is roughly a factor two. Since many
    # computers can parallellize over more than two threads, counting first --
    # which enables parallelization -- should still result in a net speed up.
    n_edge = edge_coords.shape[0]
    counts = np.empty(n_edge + 1, dtype=IntDType)
    int_dummy = np.empty((0,), dtype=IntDType)
    float_dummy = np.empty((0, 0, 0), dtype=FloatDType)
    counts[0] = 0
    # First run a count so we can allocate afterwards
    for i in nb.prange(n_edge):  # pylint: disable=not-an-iterable
        a = as_point(edge_coords[i, 0])
        b = as_point(edge_coords[i, 1])
        counts[i + 1] = locate_edge(a, b, tree, int_dummy, float_dummy, False)

    # Run a cumulative sum
    total = 0
    for i in range(1, n_edge + 1):
        total += counts[i]
        counts[i] = total

    # Now allocate appropriately
    ii = np.empty(total, dtype=IntDType)
    jj = np.empty(total, dtype=IntDType)
    xy = np.empty((total, 2, 2), dtype=FloatDType)
    for i in nb.prange(n_edge):  # pylint: disable=not-an-iterable
        start = counts[i]
        end = counts[i + 1]
        ii[start:end] = i
        indices = jj[start:end]
        intersections = xy[start:end]
        a = as_point(edge_coords[i, 0])
        b = as_point(edge_coords[i, 1])
        locate_edge(a, b, tree, indices, intersections, True)

    return ii, jj, xy


@nb.njit(cache=True)
def collect_node_bounds(tree: CellTreeData) -> FloatArray:
    # Allocate output array.
    # Per row: xmin, xmax, ymin, ymax
    node_bounds = np.empty((len(tree.nodes), 4), dtype=FloatDType)
    # Set bounds of the first node.
    node_bounds[0, 0] = tree.bbox[0]
    node_bounds[0, 1] = tree.bbox[1]
    node_bounds[0, 2] = tree.bbox[2]
    node_bounds[0, 3] = tree.bbox[3]

    stack = allocate_stack()
    parent_stack = allocate_stack()
    side_stack = allocate_stack()

    # Right child
    stack[0] = 2
    parent_stack[0] = 0
    side_stack[0] = 0
    # Left child
    stack[1] = 1
    parent_stack[1] = 0
    side_stack[1] = 1
    # Stack size starts at two.
    size = 2

    while size > 0:
        # Collect from stacks
        # Sizes are synchronized.
        parent_index, _ = pop(parent_stack, size)
        side, _ = pop(side_stack, size)
        node_index, size = pop(stack, size)

        parent = tree.nodes[parent_index]
        bbox = node_bounds[parent_index]
        dim = 1 if parent["dim"] else 0

        # Set parent bounding box first.
        # Then place the single new value for the child.
        node_bounds[node_index, 0] = bbox[0]
        node_bounds[node_index, 1] = bbox[1]
        node_bounds[node_index, 2] = bbox[2]
        node_bounds[node_index, 3] = bbox[3]

        if side:
            bound = parent["Lmax"]
        else:
            bound = parent["Rmin"]

        node_bounds[node_index, dim * 2 + side] = bound

        node = tree.nodes[node_index]
        if node["child"] == -1:
            continue

        left_child = node["child"]
        right_child = left_child + 1

        # Right child
        push(parent_stack, node_index, size)
        push(side_stack, 0, size)
        size = push(stack, right_child, size)

        # Left child
        push(parent_stack, node_index, size)
        push(side_stack, 1, size)
        size = push(stack, left_child, size)

    return node_bounds


@nb.njit(cache=True)
def validate_node_bounds(tree: CellTreeData, node_bounds: FloatArray) -> BoolArray:
    """
    Traverse the tree. Check whether all children are contained in the bounding
    box.

    For the leaf nodes, check whether the bounding boxes are contained.
    """
    node_validity = np.full(len(tree.nodes), False, dtype=np.bool_)
    stack = allocate_stack()
    stack[0] = 0
    size = 1

    while size > 0:
        node_index, size = pop(stack, size)
        bbox = as_box(node_bounds[node_index])
        node = tree.nodes[node_index]

        # Check if it's a leaf:
        if node["child"] == -1:
            valid = True
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                leaf_box = as_box(tree.bb_coords[bbox_index])
                valid = valid and box_contained(leaf_box, bbox)
            node_validity[node_index] = valid
            continue

        left_child = node["child"]
        right_child = left_child + 1
        left_box = as_box(node_bounds[left_child])
        right_box = as_box(node_bounds[right_child])
        node_validity[node_index] = box_contained(left_box, bbox) and box_contained(
            right_box, bbox
        )

        size = push(stack, right_child, size)
        size = push(stack, left_child, size)

    return node_validity



"""
Compute the Wachspress Barycentric ccoordinate for a convex polygon. These can
be directly used for barycentric interpolation.
"""

import numba as nb
import numpy as np

from numba_celltree.constants import (
    PARALLEL,
    TOLERANCE_ON_EDGE,
    FloatArray,
    FloatDType,
    IntArray,
)
from numba_celltree.geometry_utils import (
    Point,
    as_point,
    copy_vertices,
    cross_product,
    dot_product,
    to_vector,
)


@nb.njit(inline="always")
def interp_edge_case(a, U, p, weights, i, j):
    # For the edge case, find the linear interpolation weight between the
    # vertices.
    weights[:] = 0
    V = to_vector(a, p)
    w = np.sqrt(dot_product(V, V)) / np.sqrt(dot_product(U, U))
    weights[i] = 1.0 - w
    weights[j] = w
    return


@nb.njit
def compute_weights(polygon: FloatArray, p: Point, weights: FloatArray) -> None:
    n = len(polygon)
    w_sum = 0.0

    # Initial iteration
    a = as_point(polygon[-1])
    b = as_point(polygon[0])
    U = to_vector(a, b)
    V = to_vector(a, p)
    Ai = abs(cross_product(U, V))
    if Ai < TOLERANCE_ON_EDGE:
        # Note: weights may be differently sized than polygon! Hence n-1
        # instead of -1.
        interp_edge_case(a, U, p, weights, n - 1, 0)
        return

    for i in range(n):
        i_next = (i + 1) % n
        c = as_point(polygon[i_next])

        V = to_vector(a, c)
        Ci = abs(cross_product(U, V))

        U = to_vector(b, p)
        V = to_vector(b, c)
        Aj = abs(cross_product(U, V))

        if Aj < TOLERANCE_ON_EDGE:
            interp_edge_case(b, V, p, weights, i, i_next)
            return

        w = 2 * Ci / (Ai * Aj)
        weights[i] = w
        w_sum += w

        # Setup next iteration
        a = b
        b = c
        U = to_vector(a, b)
        Ai = Aj

    # normalize weights
    for i in range(n):
        weights[i] /= w_sum

    return


@nb.njit(parallel=PARALLEL, cache=True)
def barycentric_wachspress_weights(
    points: FloatArray,
    face_indices: IntArray,
    faces: IntArray,
    vertices: FloatArray,
) -> FloatArray:
    n_points = len(points)
    n_max_vert = faces.shape[1]
    weights = np.zeros((n_points, n_max_vert), dtype=FloatDType)
    for i in nb.prange(n_points):
        face_index = face_indices[i]
        if face_index == -1:
            continue
        face = faces[face_index]
        polygon = copy_vertices(vertices, face)
        point = as_point(points[i])
        compute_weights(polygon, point, weights[i])
    return weights



import numpy as np
import pytest

from numba_celltree.algorithms import barycentric_triangle as bt
from numba_celltree.algorithms import barycentric_wachspress as bwp
from numba_celltree.constants import Point, Triangle, Vector


def test_interp_edge_case():
    def compute(a, U, p):
        weights = np.zeros(4)
        bwp.interp_edge_case(a, U, p, weights, 0, 1)
        return weights

    a = Point(0.0, 0.0)
    U = Vector(1.0, 0.0)
    assert np.allclose(compute(a, U, Point(0.0, 0.0)), [1.0, 0.0, 0.0, 0.0])
    assert np.allclose(compute(a, U, Point(1.0, 0.0)), [0.0, 1.0, 0.0, 0.0])
    assert np.allclose(compute(a, U, Point(0.25, 0.0)), [0.75, 0.25, 0.0, 0.0])
    assert np.allclose(compute(a, U, Point(0.75, 0.0)), [0.25, 0.75, 0.0, 0.0])


def test_compute_weights_triangle():
    def compute(triangle, point):
        weights = np.zeros(3)
        bt.compute_weights(triangle, point, weights)
        return weights

    triangle = Triangle(
        Point(0.0, 0.0),
        Point(1.0, 0.0),
        Point(1.0, 1.0),
    )

    # Test for the vertices
    assert np.allclose(compute(triangle, Point(0.0, 0.0)), [1.0, 0.0, 0.0])
    assert np.allclose(compute(triangle, Point(1.0, 0.0)), [0.0, 1.0, 0.0])
    assert np.allclose(compute(triangle, Point(1.0, 1.0)), [0.0, 0.0, 1.0])

    # Test halfway edges
    assert np.allclose(compute(triangle, Point(0.5, 0.0)), [0.5, 0.5, 0.0])
    assert np.allclose(compute(triangle, Point(1.0, 0.5)), [0.0, 0.5, 0.5])
    assert np.allclose(compute(triangle, Point(0.5, 0.5)), [0.5, 0.0, 0.5])


@pytest.mark.parametrize(
    "barycentric_weights",
    [bt.barycentric_triangle_weights, bwp.barycentric_wachspress_weights],
)
def test_barycentric_triangle_weights(barycentric_weights):
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.5, 0.0],
            [1.0, 0.5],
            [0.5, 0.5],
            [2.0, 2.0],
        ]
    )
    vertices = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]
    )
    faces = np.array(
        [
            [0, 1, 2],
        ]
    )
    face_indices = np.array([0, 0, 0, 0, 0, 0, -1])

    expected = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.5, 0.5, 0.0],
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.0, 0.0, 0.0],
        ]
    )
    actual = barycentric_weights(points, face_indices, faces, vertices)
    assert np.allclose(actual, expected)


def test_compute_weights_wachspress():
    def compute(polygon, point):
        weights = np.zeros(4)
        bwp.compute_weights(polygon, point, weights)
        return weights

    polygon = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    assert np.allclose(compute(polygon, Point(0.0, 0.0)), [1.0, 0.0, 0.0, 0.0])
    assert np.allclose(compute(polygon, Point(1.0, 0.0)), [0.0, 1.0, 0.0, 0.0])
    assert np.allclose(compute(polygon, Point(1.0, 1.0)), [0.0, 0.0, 1.0, 0.0])
    assert np.allclose(compute(polygon, Point(0.0, 1.0)), [0.0, 0.0, 0.0, 1.0])
    assert np.allclose(compute(polygon, Point(0.25, 0.0)), [0.75, 0.25, 0.0, 0.0])
    assert np.allclose(compute(polygon, Point(0.25, 1.0)), [0.0, 0.0, 0.25, 0.75])
    assert np.allclose(compute(polygon, Point(0.5, 0.5)), [0.25, 0.25, 0.25, 0.25])


def test_barycentric_wachspress_weights():
    vertices = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    faces = np.array(
        [
            [0, 1, 2, 3],
        ]
    )
    face_indices = np.array([0, 0, 0, 0, 0, 0, 0, -1])
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.25, 0.0],
            [0.25, 1.0],
            [0.5, 0.5],
            [2.0, 2.0],
        ]
    )
    expected = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
            [0.75, 0.25, 0.0, 0.0],
            [0.0, 0.0, 0.25, 0.75],
            [0.25, 0.25, 0.25, 0.25],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    actual = bwp.barycentric_wachspress_weights(points, face_indices, faces, vertices)
    assert np.allclose(actual, expected)



import pathlib
import shutil

import numpy as np
import pytest

from numba_celltree import CellTree2d, demo


@pytest.fixture
def datadir(tmpdir, request):
    data = pathlib.Path(__file__).parent / "data"
    shutil.copy(data / "triangles.txt", tmpdir / "triangles.txt")
    shutil.copy(data / "xy.txt", tmpdir / "xy.txt")
    shutil.copy(data / "voronoi.txt", tmpdir / "voronoi.txt")
    shutil.copy(data / "voronoi_xy.txt", tmpdir / "voronoi_xy.txt")
    return tmpdir


# some very simple test data:
# two triangles:
nodes2 = [
    [0.0, 0.0],
    [2.0, 0.0],
    [1.0, 2.0],
    [3.0, 2.0],
]

faces2 = [
    [0, 1, 2],
    [1, 3, 2],
]
fill_value = -1
nodes = np.array(nodes2, dtype=np.float64)
faces = np.array(faces2, dtype=np.intc)


nodes21 = [
    (5, 1),
    (10, 1),
    (3, 3),
    (7, 3),
    (9, 4),
    (12, 4),
    (5, 5),
    (3, 7),
    (5, 7),
    (7, 7),
    (9, 7),
    (11, 7),
    (5, 9),
    (8, 9),
    (11, 9),
    (9, 11),
    (11, 11),
    (7, 13),
    (9, 13),
    (7, 15),
]

faces21 = [
    (0, 1, 3),
    (0, 2, 6),
    (0, 3, 6),
    (1, 3, 4),
    (1, 4, 5),
    (2, 6, 7),
    (6, 7, 8),
    (7, 8, 12),
    (6, 8, 9),
    (8, 9, 12),
    (9, 12, 13),
    (4, 5, 11),
    (4, 10, 11),
    (9, 10, 13),
    (10, 11, 14),
    (10, 13, 14),
    (13, 14, 15),
    (14, 15, 16),
    (15, 16, 18),
    (15, 17, 18),
    (17, 18, 19),
]


def disk():
    vertices, faces = demo.generate_disk(5, 5)
    centroids = vertices[faces].mean(axis=1)
    rsquared = (centroids[:, 0] - -1.0) ** 2 + (centroids[:, 1] - -1.0) ** 2
    order = np.argsort(rsquared)
    faces = faces[order]
    return vertices, faces


def test_init():
    """Can a tree be initialized?"""
    # with defaults
    CellTree2d(nodes, faces, fill_value)
    # with everything specified
    CellTree2d(nodes, faces, fill_value, n_buckets=2, cells_per_leaf=1)
    # with n_buckets
    CellTree2d(nodes, faces, fill_value, n_buckets=4)
    # with cells_per_leaf
    CellTree2d(nodes, faces, fill_value, cells_per_leaf=2)


def test_init_larger_mesh(datadir):
    # This mesh is large enough so that a bucket will get split during
    # construction.
    nodes = np.loadtxt(datadir / "xy.txt", dtype=float)
    faces = np.loadtxt(datadir / "triangles.txt", dtype=int)
    CellTree2d(nodes, faces, fill_value, n_buckets=2)


def test_lists():
    """Python lists should get converted to numpy arrays"""
    CellTree2d(nodes2, faces2, fill_value)


def test_types():
    """It should auto-cast the types to the right types for you"""
    nodes = np.array(nodes2, dtype=np.float32)
    faces = np.array(faces2, dtype=np.int32)
    CellTree2d(nodes, faces, fill_value)


def test_fill_value_conversion():
    faces = np.array([[0, 1, 2, -999], [1, 3, 2, -999]])
    tree = CellTree2d(nodes, faces, -999)
    assert tree.faces[0, -1] == -1
    assert tree.faces[1, -1] == -1


def test_shape_errors():
    faces = [0, 1, 2, 1, 3, 2]
    nodes = [(1, 2, 3), (3, 4, 5), (4, 5, 6)]
    box_coords = np.array([0.0, 1.0, 2.0, 3.0])
    edge_coords = np.array([[0.0, 1.0], [2.0, 3.0]])
    with pytest.raises(ValueError):
        CellTree2d(nodes, faces2, -1)
    with pytest.raises(ValueError):
        CellTree2d(nodes2, faces, -1)

    tree = CellTree2d(nodes2, faces2, -1)
    with pytest.raises(ValueError):
        tree.locate_points(nodes)
    with pytest.raises(ValueError):
        tree.intersect_faces(nodes, faces2, -1)
    with pytest.raises(ValueError):
        tree.intersect_faces(faces, nodes2, -1)
    with pytest.raises(ValueError):
        tree.locate_boxes(box_coords)
    with pytest.raises(ValueError):
        tree.intersect_boxes(box_coords)
    with pytest.raises(ValueError):
        tree.intersect_edges(edge_coords)


def test_bounds_errors():
    with pytest.raises(ValueError):
        CellTree2d(nodes, faces, fill_value, cells_per_leaf=-1)

    with pytest.raises(ValueError):
        CellTree2d(nodes, faces, fill_value, n_buckets=0)


def test_triangle_lookup():
    tree = CellTree2d(nodes, faces, fill_value)
    point = np.array(
        [
            [1.0, 1.0],
            [2.0, 1.0],
            [-1.0, 1.0],
        ]
    )  # in triangle 1
    result = tree.locate_points(point)
    expected = np.array([0, 1, -1])
    assert np.array_equal(result, expected)


def test_poly_lookup():
    # A simple quad grid
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [0.0, 2.0],  # 1
            [2.0, 0.0],  # 2
            [2.0, 2.0],  # 3
            [4.0, 0.0],  # 4
            [4.0, 2.0],  # 5
            [6.0, 0.0],  # 6
            [6.0, 2.0],  # 7
            [0.0, 4.0],  # 8
            [2.0, 4.0],  # 9
            [4.0, 4.0],  # 10
            [6.0, 4.0],  # 11
        ]
    )

    # quads
    faces1 = np.array(
        [
            [0, 2, 3, 1],
            [4, 6, 7, 5],
        ],
        dtype=np.intc,
    )
    # Pentas
    faces2 = np.array(
        [
            [0, 8, 9, 5, 2],
            [9, 11, 6, 2, 5],
        ],
        dtype=np.intc,
    )

    tree1 = CellTree2d(nodes, faces1, fill_value, n_buckets=2, cells_per_leaf=1)
    point = np.array(
        [
            [1.0, 1.0],
            [5.0, 1.0],
            [-1.0, 1.0],
        ]
    )
    result = tree1.locate_points(point)
    expected = np.array([0, 1, -1])
    assert np.array_equal(result, expected)

    tree2 = CellTree2d(nodes, faces2, fill_value, n_buckets=2, cells_per_leaf=1)
    point = np.array(
        [
            [1.0, 2.0],
            [5.0, 2.0],
            [-1.0, 2.0],
        ]
    )
    result = tree2.locate_points(point)
    expected = np.array([0, 1, -1])
    assert np.array_equal(result, expected)


def test_multi_poly_lookup():
    # A simple quad grid
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [0.0, 2.0],  # 1
            [2.0, 0.0],  # 2
            [2.0, 2.0],  # 3
            [4.0, 0.0],  # 4
            [4.0, 2.0],  # 5
            [6.0, 0.0],  # 6
            [6.0, 2.0],  # 7
            [0.0, 4.0],  # 8
            [2.0, 4.0],  # 9
            [4.0, 4.0],  # 10
            [6.0, 4.0],  # 11
        ]
    )

    faces = np.array(
        [[0, 8, 9, 5, 2], [9, 11, 7, 5, -1], [4, 7, 6, -1, -1]], dtype=np.intc
    )
    tree = CellTree2d(nodes, faces, fill_value, n_buckets=2, cells_per_leaf=1)
    point = np.array(
        [
            [1.0, 1.0],
            [5.0, 0.5],
            [5.0, 3.0],
            [-1.0, 1.0],
        ]
    )
    result = tree.locate_points(point)
    expected = np.array([0, 2, 1, -1])
    assert np.array_equal(result, expected)


def test_multipoint():
    tree = CellTree2d(nodes21, faces21, fill_value)
    points = [
        (4.2, 3.0),
        (7.7, 13.5),
        (3.4, 7.000000001),
        (7.0, 5.0),  # out of bounds points
        (8.66, 10.99),
        (7.3, 0.74),
        (2.5, 5.5),
        (9.8, 12.3),
    ]
    expected = (1, 20, 7, -1, -1, -1, -1, -1)
    actual = tree.locate_points(points)
    assert np.array_equal(actual, expected)


def test_box_lookup():
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [0.0, 2.0],  # 1
            [2.0, 0.0],  # 2
            [2.0, 2.0],  # 3
            [4.0, 0.0],  # 4
            [4.0, 2.0],  # 5
            [6.0, 0.0],  # 6
            [6.0, 2.0],  # 7
            [0.0, 4.0],  # 8
            [2.0, 4.0],  # 9
            [4.0, 4.0],  # 10
            [6.0, 4.0],  # 11
        ]
    )

    faces = np.array(
        [[0, 8, 9, 5, 2], [9, 11, 7, 5, -1], [4, 7, 6, -1, -1]], dtype=np.intc
    )
    tree = CellTree2d(nodes, faces, fill_value, n_buckets=2, cells_per_leaf=1)
    box_coords = np.array(
        [
            [1.0, 2.0, 1.0, 2.0],  # in face 0
            [4.0, 5.0, 0.0, 1.0],  # in face 2
            [4.0, 5.0, 2.0, 3.0],  # in face 1
            [-1.0, 0.0, 0.0, 4.0],  # out of bounds x
            [6.0, 8.0, 0.0, 4.0],  # out of bounds x
            [0.0, 6.0, -1.0, 0.0],  # out of bounds y
            [0.0, 6.0, 4.0, 5.0],  # out of bounds y
        ]
    )
    actual_i, actual_j = tree.locate_boxes(box_coords)
    expected_i = np.array([0, 1, 2])
    expected_j = np.array([0, 2, 1])
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)

    actual_i, actual_j, _ = tree.intersect_boxes(box_coords)
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)


def test_edge_lookup():
    nodes = np.array(
        [
            [0.0, 0.0],  # 0
            [0.0, 2.0],  # 1
            [2.0, 0.0],  # 2
            [2.0, 2.0],  # 3
            [4.0, 0.0],  # 4
            [4.0, 2.0],  # 5
            [6.0, 0.0],  # 6
            [6.0, 2.0],  # 7
            [0.0, 4.0],  # 8
            [2.0, 4.0],  # 9
            [4.0, 4.0],  # 10
            [6.0, 4.0],  # 11
        ]
    )

    faces = np.array(
        [[0, 8, 9, 5, 2], [9, 11, 7, 5, -1], [4, 7, 6, -1, -1]], dtype=np.intc
    )
    tree = CellTree2d(nodes, faces, fill_value, n_buckets=2, cells_per_leaf=1)
    edge_coords = np.array(
        [
            [[1.0, 1.0], [2.0, 2.0]],  # 0
            [[4.0, 3.0], [5.0, 4.0]],  # 1
            [[5.0, 0.0], [6.0, 1.0]],  # 2
            [[-2.0, -1.0], [0.0, 1.0]],  # out of bounds
            [[-2.0, -1.0], [-2.0, -1.0]],  # out of bbox
        ]
    )
    actual_i, actual_j, intersections = tree.intersect_edges(edge_coords)
    expected_i = np.array([0, 1, 2])
    expected_j = np.array([0, 1, 2])
    expected_intersections = edge_coords[:3]
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)
    assert np.array_equal(intersections, expected_intersections)

    # Flip edge orientation
    actual_i, actual_j, intersections = tree.intersect_edges(edge_coords[:, ::-1])
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)
    assert np.array_equal(intersections, expected_intersections[:, ::-1])


def test_example_material():
    # Note: the concatenation of lists to get 1D arrays is purely to keep black
    # from formatting everything into very long 1-element columns.
    vertices, faces = disk()
    vertices += 1.0
    vertices *= 5.0

    tree = CellTree2d(vertices, faces, -1)
    points = np.array(
        [
            [-5.0, 1.0],
            [4.5, 2.5],
            [6.5, 4.5],
        ]
    )
    expected = [-1, 24, 63]
    assert np.array_equal(tree.locate_points(points), expected)

    box_coords = np.array(
        [
            [4.0, 8.0, 4.0, 6.0],
            [0.0, 8.0, 8.0, 10.0],
            [10.0, 13.0, 2.0, 8.0],
        ]
    )
    expected_i = np.concatenate(
        [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
    )
    expected_j = np.concatenate(
        [
            [107, 109, 99, 106, 95, 93, 92, 80, 62, 73, 86, 97, 81, 79, 65, 63],
            [76, 57, 44, 59, 54, 58, 50, 45, 43, 31, 41, 36, 38, 27, 123, 112],
            [120, 117, 121, 108, 110, 115, 116, 101, 104, 103, 111, 88, 98],
            [90, 100, 72, 84, 85, 66, 75, 69],
        ]
    )
    i, j = tree.locate_boxes(box_coords)
    assert np.array_equal(i, expected_i)
    assert np.array_equal(j, expected_j)

    triangle_vertices = np.array(
        [
            [5.0, 3.0],
            [7.0, 3.0],
            [7.0, 5.0],
            [0.0, 6.0],
            [4.0, 4.0],
            [6.0, 10.0],
        ]
    )
    triangles = np.array(
        [
            [0, 1, 2],
            [3, 4, 5],
        ]
    )

    expected_i = np.concatenate(
        [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [1],
        ]
    )
    expected_j = np.concatenate(
        [
            [81, 79, 61, 65, 63, 76, 44, 59, 48, 40, 115, 116, 101],
            [104, 91, 103, 111, 88, 98, 83, 72, 84, 68, 74, 54, 58, 55],
            [67, 47, 35, 50, 66, 46, 53, 26, 34, 49, 56, 29, 37, 43],
            [36, 38, 27, 30],
        ]
    )

    i, j, _ = tree.intersect_faces(triangle_vertices, triangles, -1)
    assert np.array_equal(i, expected_i)
    assert np.array_equal(j, expected_j)

    edge_coords = np.array(
        [
            [[0.0, 0.0], [10.0, 10.0]],
            [[0.0, 10.0], [10.0, 0.0]],
        ]
    )
    expected_i = np.concatenate(
        [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
    )
    expected_j = np.concatenate(
        [
            [124, 112, 120, 108, 93, 96, 80, 62, 73, 43, 36, 27, 16],
            [21, 11, 2, 6, 0, 70, 71, 77, 64, 61, 65, 63, 57, 59],
            [72, 54, 58, 55, 67, 50, 66, 75, 69],
        ]
    )
    i, j, _ = tree.intersect_edges(edge_coords)
    assert np.array_equal(i, expected_i)
    assert np.array_equal(j, expected_j)


def test_compute_barycentric_weights_triangles():
    tree = CellTree2d(nodes, faces, fill_value)
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 1.0],
        ]
    )
    face_indices, weights = tree.compute_barycentric_weights(points)

    expected_indices = np.array([0, 0, 1])
    expected_weights = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.25, 0.25, 0.5],
            [0.5, 0.25, 0.25],
        ]
    )
    assert np.array_equal(face_indices, expected_indices)
    assert np.allclose(weights, expected_weights)


def test_compute_barycentric_weights_triangle_quads():
    nodes = np.array(
        [
            [0.0, 0.0],
            [2.0, 0.0],
            [2.0, 2.0],
            [0.0, 2.0],
            [4.0, 0.0],
            [4.0, 4.0],
        ]
    )
    faces = np.array(
        [
            [0, 1, 2, 3],
            [1, 4, 5, 2],
        ]
    )
    fill_value = -1
    tree = CellTree2d(nodes, faces, fill_value)
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 1.0],
        ]
    )
    face_indices, weights = tree.compute_barycentric_weights(points)

    expected_indices = np.array([0, 0, 1])
    expected_weights = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.25, 0.25, 0.25, 0.25],
            [0.5, 0.0, 0.0, 0.5],
        ]
    )
    assert np.array_equal(face_indices, expected_indices)
    assert np.allclose(weights, expected_weights)


def test_node_bounds(datadir):
    nodes = np.loadtxt(datadir / "voronoi_xy.txt", dtype=float)
    faces = np.loadtxt(datadir / "voronoi.txt", dtype=int)
    tree = CellTree2d(nodes, faces, fill_value)
    node_bounds = tree.node_bounds

    n_node = len(tree.celltree_data.nodes)
    xmin, xmax, ymin, ymax = tree.bbox
    assert node_bounds.shape == (n_node, 4)
    assert (node_bounds[:, 0] >= xmin).all()
    assert (node_bounds[:, 1] <= xmax).all()
    assert (node_bounds[:, 2] >= ymin).all()
    assert (node_bounds[:, 3] <= ymax).all()


def test_node_validity(datadir):
    nodes = np.loadtxt(datadir / "voronoi_xy.txt", dtype=float)
    faces = np.loadtxt(datadir / "voronoi.txt", dtype=int)
    tree = CellTree2d(nodes, faces, fill_value)

    assert tree.validate_node_bounds().all()

    # Introduce a failure as described in:
    # https://github.com/Deltares/numba_celltree/issues/2

    tree.nodes[71]["Lmax"] = -0.02319655
    validity = tree.validate_node_bounds()
    assert not validity[73]


def test_find_centroids(datadir):
    nodes = np.loadtxt(datadir / "xy.txt", dtype=float)
    faces = np.loadtxt(datadir / "triangles.txt", dtype=int)
    centroids = nodes[faces].mean(axis=1)
    tree = CellTree2d(nodes, faces, fill_value)
    expected = np.arange(len(centroids))
    actual = tree.locate_points(centroids)
    assert np.array_equal(expected, actual)


def test_to_dict_of_lists(datadir):
    nodes = np.loadtxt(datadir / "xy.txt", dtype=float)
    faces = np.loadtxt(datadir / "triangles.txt", dtype=int)
    tree = CellTree2d(nodes, faces, fill_value, n_buckets=4)
    d = tree.to_dict_of_lists()

    assert isinstance(d, dict)
    assert list(d.keys()) == list(range(len(tree.celltree_data.nodes)))
    assert max(len(v) for v in d.values()) == 2



from typing import Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import (
    FLOAT_MAX,
    FLOAT_MIN,
    PARALLEL,
    BoolArray,
    FloatArray,
    IntArray,
)
from numba_celltree.geometry_utils import Vector, as_point, copy_vertices, dot_product


@nb.njit(inline="always")
def extrema_projected(
    norm: Vector, polygon: FloatArray, length: int
) -> Tuple[float, float]:
    min_proj = FLOAT_MAX
    max_proj = FLOAT_MIN
    for i in range(length):
        proj = dot_product(as_point(polygon[i]), norm)
        min_proj = min(min_proj, proj)
        max_proj = max(max_proj, proj)
    return min_proj, max_proj


@nb.njit(inline="always")
def is_separating_axis(
    norm: Vector, a: FloatArray, b: FloatArray, length_a: int, length_b: int
) -> bool:
    mina, maxa = extrema_projected(norm, a, length_a)
    minb, maxb = extrema_projected(norm, b, length_b)
    if maxa > minb and maxb > mina:
        return False
    else:
        return True


@nb.njit(inline="always")
def separating_axes(a: FloatArray, b: FloatArray) -> bool:
    length_a = len(a)
    length_b = len(b)
    p = as_point(a[length_a - 1])
    for i in range(length_a):
        q = as_point(a[i])
        norm = Vector(p.y - q.y, q.x - p.x)
        p = q
        if norm.x == 0.0 and norm.y == 0.0:
            continue
        if is_separating_axis(norm, a, b, length_a, length_b):
            return False
    return True


@nb.njit(parallel=PARALLEL, cache=True)
def polygons_intersect(
    vertices_a: FloatArray,
    vertices_b: FloatArray,
    faces_a: IntArray,
    faces_b: IntArray,
    indices_a: IntArray,
    indices_b: IntArray,
) -> BoolArray:
    n_shortlist = indices_a.size
    intersects = np.empty(n_shortlist, dtype=np.bool_)
    for i in nb.prange(n_shortlist):
        face_a = faces_a[indices_a[i]]
        face_b = faces_b[indices_b[i]]
        a = copy_vertices(vertices_a, face_a)
        b = copy_vertices(vertices_b, face_b)
        intersects[i] = separating_axes(a, b) and separating_axes(b, a)
    return intersects



import numpy as np

from numba_celltree.algorithms.separating_axis import (
    polygons_intersect,
    separating_axes,
)


def test_triangles_intersect():
    # Identity
    a = np.array(
        [
            [0.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0],
        ]
    )
    b = a
    assert separating_axes(a, b)
    assert separating_axes(b, a)

    # No overlap
    b = np.array(
        [
            [2.0, 0.0],
            [3.0, 1.0],
            [2.0, 1.0],
        ]
    )
    assert not separating_axes(a, b)
    assert not separating_axes(b, a)

    # Touching: does not qualify
    b = np.array(
        [
            [1.0, 0.0],
            [2.0, 1.0],
            [1.0, 1.0],
        ]
    )
    assert not separating_axes(a, b)
    assert not separating_axes(b, a)

    # One inside the other
    a = np.array(
        [
            [0.0, 0.0],
            [4.0, 0.0],
            [0.0, 4.0],
        ]
    )
    b = np.array(
        [
            [1.0, 1.0],
            [2.0, 1.0],
            [1.0, 2.0],
        ]
    )
    assert separating_axes(a, b)
    assert separating_axes(b, a)

    # Mirrored
    a = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 1.0],
        ]
    )
    b = np.array(
        [
            [1.0, 0.0],
            [2.0, 0.0],
            [2.0, 1.0],
        ]
    )
    assert not separating_axes(a, b)
    assert not separating_axes(b, a)

    # This is a case which requires testing both (a, b) and (b, a) to determine intersection
    # two edges of a do separate b
    # no edges of b separate a
    # => no intersection
    a = np.array(
        [
            [5.0, 3.0],
            [7.0, 3.0],
            [7.0, 5.0],
        ]
    )
    b = np.array(
        [
            [6.9, 5.6],
            [8.0, 4.75],
            [7.85, 5.9],
        ]
    )
    assert separating_axes(a, b)
    assert not separating_axes(b, a)
    assert not (separating_axes(a, b) and separating_axes(b, a))


def test_box_triangle():
    a = np.array(
        [
            [0.0, 0.0],
            [2.0, 0.0],
            [2.0, 2.0],
            [0.0, 2.0],
        ]
    )
    b = np.array(
        [
            [1.0, 1.0],
            [3.0, 0.0],
            [3.0, 1.0],
        ]
    )
    assert separating_axes(a, b)
    assert separating_axes(b, a)

    # Touching
    b = np.array(
        [
            [2.0, 1.0],
            [3.0, 0.0],
            [3.0, 1.0],
        ]
    )
    assert not separating_axes(a, b)
    assert not separating_axes(b, a)

    # Inside
    b = np.array(
        [
            [0.25, 0.25],
            [0.75, 0.75],
            [0.6, 0.6],
        ]
    )
    assert separating_axes(a, b)
    assert separating_axes(b, a)


def test_polygons_intersect():
    vertices_a = np.array(
        [
            [0.0, 0.0],
            [2.0, 0.0],
            [2.0, 2.0],
            [0.0, 2.0],
        ]
    )
    faces_a = np.array(
        [
            [0, 1, 2, 3],
        ]
    )
    vertices_b = np.array(
        [
            [0.25, 0.25],
            [0.75, 0.75],
            [0.6, 0.6],
            [2.0, 1.0],
            [3.0, 0.0],
            [3.0, 1.0],
        ]
    )
    faces_b = np.array(
        [
            [0, 1, 2, -1],
            [3, 4, 5, -1],
        ]
    )
    indices_a = np.array([0, 0])
    indices_b = np.array([0, 1])

    actual = polygons_intersect(
        vertices_a, vertices_b, faces_a, faces_b, indices_a, indices_b
    )
    expected = np.array([True, False])
    assert np.array_equal(actual, expected)


def test_triangles_intersect_hanging_nodes():
    # Identity
    a = np.array(
        [
            [0.0, 0.0],
            [0.5, 0.5],
            [0.0, 1.0],
            [1.0, 1.0],
        ]
    )
    b = a
    assert separating_axes(a, b)
    assert separating_axes(b, a)

    # No overlap
    b = np.array(
        [
            [2.0, 0.0],
            [3.0, 1.0],
            [2.5, 1.0],
            [2.0, 1.0],
        ]
    )
    assert not separating_axes(a, b)
    assert not separating_axes(b, a)

    # Repeat node
    a = np.array(
        [
            [0.0, 0.0],
            [0.5, 0.5],
            [0.5, 0.5],
            [0.0, 1.0],
            [1.0, 1.0],
        ]
    )
    b = a
    assert separating_axes(a, b)
    assert separating_axes(b, a)



from typing import Sequence, Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import (
    FILL_VALUE,
    NDIM,
    PARALLEL,
    TOLERANCE_ON_EDGE,
    Box,
    FloatArray,
    FloatDType,
    IntArray,
    Point,
    Triangle,
    Vector,
)
from numba_celltree.utils import allocate_box_polygon, allocate_polygon


@nb.njit(inline="always")
def to_vector(a: Point, b: Point) -> Vector:
    return Vector(b.x - a.x, b.y - a.y)


@nb.njit(inline="always")
def as_point(a: FloatArray) -> Point:
    return Point(a[0], a[1])


@nb.njit(inline="always")
def as_box(arr: FloatArray) -> Box:
    return Box(
        arr[0],
        arr[1],
        arr[2],
        arr[3],
    )


@nb.njit(inline="always")
def as_triangle(vertices: FloatArray, face: IntArray) -> Triangle:
    return Triangle(
        as_point(vertices[face[0]]),
        as_point(vertices[face[1]]),
        as_point(vertices[face[2]]),
    )


@nb.njit(inline="always")
def to_point(t: float, a: Point, V: Vector) -> Point:
    return Point(a.x + t * V.x, a.y + t * V.y)


@nb.njit(inline="always")
def cross_product(u: Vector, v: Vector) -> float:
    return u.x * v.y - u.y * v.x


@nb.njit(inline="always")
def dot_product(u: Vector, v: Vector) -> float:
    return u.x * v.x + u.y * v.y


@nb.njit(inline="always")
def polygon_length(face: IntArray) -> int:
    # A minimal polygon is a triangle
    n = len(face)
    for i in range(3, n):
        if face[i] == FILL_VALUE:
            return i
    return n


@nb.njit(inline="always")
def polygon_area(polygon: Sequence) -> float:
    length = len(polygon)
    area = 0.0
    a = Point(polygon[0][0], polygon[0][1])
    b = Point(polygon[1][0], polygon[1][1])
    U = to_vector(a, b)
    for i in range(2, length):
        c = Point(polygon[i][0], polygon[i][1])
        V = to_vector(c, a)
        area += abs(cross_product(U, V))
        b = c
        U = V
    return 0.5 * area


@nb.njit(inline="always")
def point_in_polygon(p: Point, poly: Sequence) -> bool:
    # Refer to: https://wrf.ecse.rpi.edu/Research/Short_Notes/pnpoly.html
    # Copyright (c) 1970-2003, Wm. Randolph Franklin
    # MIT license.
    #
    # Quote:
    # > I run a semi-infinite ray horizontally (increasing x, fixed y) out from
    # > the test point, and count how many edges it crosses. At each crossing,
    # > the ray switches between inside and outside. This is called the Jordan
    # > curve theorem.
    # >
    # > The case of the ray going thru a vertex is handled correctly via a
    # > careful selection of inequalities. Don't mess with this code unless
    # > you're familiar with the idea of Simulation of Simplicity. This pretends
    # > to shift the ray infinitesimally down so that it either clearly
    # > intersects, or clearly doesn't touch. Since this is merely a conceptual,
    # > infinitesimal, shift, it never creates an intersection that didn't exist
    # > before, and never destroys an intersection that clearly existed before.
    # >
    # > The ray is tested against each edge thus:
    # > 1. Is the point in the half-plane to the left of the extended edge? and
    # > 2. Is the point's Y coordinate within the edge's Y-range?
    # >
    # > Handling endpoints here is tricky.
    #
    # For the Simulation of Simplicity concept, see:
    # Edelsbrunner, H., & Mücke, E. P. (1990). Simulation of simplicity: a
    # technique to cope with degenerate cases in geometric algorithms. ACM
    # Transactions on Graphics (tog), 9(1), 66-104.
    #
    # In this case, this guarantees there will be no "on-edge" answers, which
    # are degenerative. For another application of simulation of simplicity,
    # see:
    # Rappoport, A. (1991). An efficient algorithm for line and polygon
    # clipping. The Visual Computer, 7(1), 19-28.
    length = len(poly)
    v0 = as_point(poly[-1])
    c = False
    for i in range(length):
        v1 = as_point(poly[i])
        # Do not split this in two conditionals: if the first conditional fails,
        # the second will not be executed in Python's (and C's) execution model.
        # This matters because the second can result in division by zero.
        if (v0.y > p.y) != (v1.y > p.y) and p.x < (
            (v1.x - v0.x) * (p.y - v0.y) / (v1.y - v0.y) + v0.x
        ):
            c = not c
        v0 = v1
    return c


@nb.njit(inline="always")
def in_bounds(p: Point, a: Point, b: Point) -> bool:
    """
    Check whether point p falls within the bounding box created by a and b
    (after we've checked the size of the cross product).

    However, we must take into account that a line may be either vertical
    (dx=0) or horizontal (dy=0) and only evaluate the non-zero value.

    If the area created by p, a, b is tiny AND p is within the bounds of a and
    b, the point lies very close to the edge.
    """
    dx = b.x - a.x
    dy = b.y - a.y
    if abs(dx) >= abs(dy):
        if dx > 0:
            return a.x <= p.x and p.x <= b.x
        return b.x <= p.x and p.x <= a.x
    else:
        if dy > 0:
            return a.y <= p.y and p.y <= b.y
        return b.y <= p.y and p.y <= a.y


@nb.njit(inline="always")
def point_in_polygon_or_on_edge(p: Point, poly: FloatArray) -> bool:
    length = len(poly)
    v0 = as_point(poly[-1])
    U = to_vector(p, v0)
    c = False
    for i in range(length):
        v1 = as_point(poly[i])
        if v1 == v0:
            continue
        V = to_vector(p, v1)
        # Compute the (twofold) area of formed by the point (p) and two
        # vertices of the polygon (v0, v1). If this area is extremely small,
        # the point is (nearly) on the edge, or it is collinear. We can test if
        # if's collinear by checking whether it falls in the bounding box of
        # points v0 and v1.
        if (abs(cross_product(U, V)) < TOLERANCE_ON_EDGE) and in_bounds(p, v0, v1):
            return True

        if (v0.y > p.y) != (v1.y > p.y) and p.x < (
            (v1.x - v0.x) * (p.y - v0.y) / (v1.y - v0.y) + v0.x
        ):
            c = not c

        v0 = v1
        U = V
    return c


@nb.njit(inline="always")
def boxes_intersect(a: Box, b: Box) -> bool:
    """
    Parameters
    ----------
    a: (xmin, xmax, ymin, ymax)
    b: (xmin, xmax, ymin, ymax)
    """
    return a.xmin < b.xmax and b.xmin < a.xmax and a.ymin < b.ymax and b.ymin < a.ymax


@nb.njit(inline="always")
def box_contained(a: Box, b: Box) -> bool:
    """
    Whether a is contained by b.

    Parameters
    ----------
    a: (xmin, xmax, ymin, ymax)
    b: (xmin, xmax, ymin, ymax)
    """
    return (
        a.xmin >= b.xmin and a.xmax <= b.xmax and a.ymin >= b.ymin and a.ymax <= b.ymax
    )


@nb.njit(inline="always")
def bounding_box(
    polygon: IntArray, vertices: FloatArray
) -> Tuple[float, float, float, float]:
    max_n_verts = len(polygon)
    first_vertex = vertices[polygon[0]]
    xmin = xmax = first_vertex[0]
    ymin = ymax = first_vertex[1]
    for i in range(1, max_n_verts):
        index = polygon[i]
        if index == FILL_VALUE:
            break
        vertex = vertices[index]
        x = vertex[0]
        y = vertex[1]
        xmin = min(xmin, x)
        xmax = max(xmax, x)
        ymin = min(ymin, y)
        ymax = max(ymax, y)
    return (xmin, xmax, ymin, ymax)


@nb.njit(cache=True)
def build_bboxes(
    faces: IntArray,
    vertices: FloatArray,
) -> FloatArray:
    # Make room for the bounding box of every polygon.
    n_polys = len(faces)
    bbox_coords = np.empty((n_polys, NDIM * 2), FloatDType)

    for i in nb.prange(n_polys):  # pylint: disable=not-an-iterable
        polygon = faces[i]
        bbox_coords[i] = bounding_box(polygon, vertices)

    return bbox_coords


@nb.njit(inline="always")
def copy_vertices(vertices: FloatArray, face: IntArray) -> FloatArray:
    length = polygon_length(face)
    out = allocate_polygon()
    for i in range(length):
        v = vertices[face[i]]
        out[i, 0] = v[0]
        out[i, 1] = v[1]
    return out[:length]


@nb.njit(inline="always")
def copy_vertices_into(
    vertices: FloatArray, face: IntArray, out: FloatArray
) -> FloatArray:
    length = polygon_length(face)
    for i in range(length):
        v = vertices[face[i]]
        out[i, 0] = v[0]
        out[i, 1] = v[1]
    return out[:length]


@nb.njit(inline="always")
def copy_box_vertices(box: Box) -> FloatArray:
    a = allocate_box_polygon()
    a[0, 0] = box.xmin
    a[0, 1] = box.ymin
    a[1, 0] = box.xmax
    a[1, 1] = box.ymin
    a[2, 0] = box.xmax
    a[2, 1] = box.ymax
    a[3, 0] = box.xmin
    a[3, 1] = box.ymax
    return a


@nb.njit(inline="always")
def point_inside_box(a: Point, box: Box):
    return box.xmin < a.x and a.x < box.xmax and box.ymin < a.y and a.y < box.ymax


@nb.njit(inline="always")
def flip(face: IntArray, length: int) -> None:
    end = length - 1
    for i in range(int(length / 2)):
        j = end - i
        face[i], face[j] = face[j], face[i]
    return


@nb.njit(parallel=PARALLEL, cache=True)
def counter_clockwise(vertices: FloatArray, faces: IntArray) -> None:
    n_face = len(faces)
    for i_face in nb.prange(n_face):
        face = faces[i_face]
        length = polygon_length(face)
        a = as_point(vertices[face[length - 2]])
        b = as_point(vertices[face[length - 1]])
        for i in range(length):
            c = as_point(vertices[face[i]])
            u = to_vector(a, b)
            v = to_vector(a, c)
            product = cross_product(u, v)
            if product == 0:
                a = b
                b = c
            elif product < 0:
                flip(face, length)
            else:
                break
    return



"""
Types and constants.

Note: it is not advisable to change either the IntDType or the FloatDType.

In case of the floats, 64-bit has much higher precision, which may be quite
relevant for geometry: more precision for edge cases, etc.

In case of the integers, Python has surprisingly (ambiguous) behavior. Since
numba requires static types, they've chosen to default to ``np.intp``:

See:

https://numba.pydata.org/numba-doc/latest/proposals/integer-typing.html

This means that setting IntDType to np.int32 on a 64-bit system will break
compilation: within this code, the size of an array (via ``.size``) will have
an integer type of ``np.intp``. If IntDType == np.int32, the BucketDType array
will expect a 32-bit integer for its index and size fields, yet receive a
64-bit integer (intp), and error during type inferencing.
"""

from typing import NamedTuple

import numba as nb
import numba.types as nbtypes
import numpy as np

IntDType = np.intp
FloatDType = np.float64
BoolArray = np.ndarray
IntArray = np.ndarray
FloatArray = np.ndarray
BucketArray = np.ndarray
NodeArray = np.ndarray


class Point(NamedTuple):
    x: float
    y: float


class Vector(NamedTuple):
    x: float
    y: float


class Interval(NamedTuple):
    xmin: float
    xmax: float


class Box(NamedTuple):
    xmin: float
    xmax: float
    ymin: float
    ymax: float


class Triangle(NamedTuple):
    a: Point
    b: Point
    c: Point


class Node(NamedTuple):
    child: IntDType
    Lmax: FloatDType
    Rmin: FloatDType
    ptr: IntDType
    size: IntDType
    dim: bool


class Bucket(NamedTuple):
    Max: FloatDType
    Min: FloatDType
    Rmin: FloatDType
    Lmax: FloatDType
    index: IntDType
    size: IntDType


class CellTreeData(NamedTuple):
    faces: IntArray
    vertices: FloatArray
    nodes: NodeArray
    bb_indices: IntArray
    bb_coords: FloatArray
    bbox: FloatArray
    cells_per_leaf: int


NodeDType = np.dtype(
    [
        # Index of left child. Right child is child + 1.
        ("child", IntDType),
        # Range of the bounding boxes inside of the node.
        ("Lmax", FloatDType),
        ("Rmin", FloatDType),
        # Index into the bounding box index array, bb_indices.
        ("ptr", IntDType),
        # Number of bounding boxes in this node.
        ("size", IntDType),
        # False = 0 = x, True = 1 = y.
        ("dim", bool),
    ]
)


BucketDType = np.dtype(
    [
        # Range of the bucket.
        ("Max", FloatDType),
        ("Min", FloatDType),
        # Range of the bounding boxes inside the bucket.
        ("Rmin", FloatDType),
        ("Lmax", FloatDType),
        # Index into the bounding box index array, bb_indices.
        ("index", IntDType),
        # Number of bounding boxes in this bucket.
        ("size", IntDType),
    ]
)

# Numba can parallellize for loops with a single keyword.
PARALLEL = True
# By default, Numba will allocate all arrays on the heap. For small (statically
# sized) arrays, this creates a large overhead. This enables stack allocated
# arrays rather than "regular" heap allocated numpy arrays. See allocate
# functions in utils.py.
STACK_ALLOCATE_STATIC_ARRAYS = True
# 2D is still rather hard-baked in, so changing this alone to 3 will NOT
# suffice to generalize it to a 3D CellTree.
NDIM = 2
MAX_N_VERTEX = 32
FILL_VALUE = -1
# Recursion in numba is somewhat slow (in case of querying), or unsupported for
# AOT-compilation when creating. We can avoid recursing by manually maintaining
# a stack, and pushing and popping. To estimate worst case, let's assume every
# leaf (node) of the tree contains only a single cell. Then the number of cells
# that can be included is given by 2 ** (depth of stack - 1).
# (int(math.ceil(math.log(MAX_N_FACE, 2))) + 1)
# This is only true for relatively balanced trees. MAX_N_FACE = int(2e9)
# results in required stack of 32. 128 suffices for some more degenerate input
# (triangulation result of complex polygon earcuts).
MAX_TREE_DEPTH = 128
# Floating point slack
TOLERANCE_ON_EDGE = 1e-9

# Derived types & constants
NumbaFloatDType = nb.from_dtype(FloatDType)
NumbaIntDType = nb.from_dtype(IntDType)
NumbaNodeDType = nb.from_dtype(NodeDType)
NumbaBucketDType = nb.from_dtype(BucketDType)

NumbaCellTreeData = nbtypes.NamedTuple(
    (
        NumbaIntDType[:, :],  # faces
        NumbaFloatDType[:, :],  # vertices
        NumbaNodeDType[:],  # nodes
        NumbaIntDType[:],  # bb_indices
        NumbaFloatDType[:, :],  # bb_coords
        NumbaFloatDType[:],  # bbox
        NumbaIntDType,  # cells_per_leaf
    ),
    CellTreeData,
)

FLOAT_MIN = np.finfo(FloatDType).min
FLOAT_MAX = np.finfo(FloatDType).max
INT_MAX = np.iinfo(IntDType).max



from numba_celltree.algorithms.barycentric_triangle import barycentric_triangle_weights
from numba_celltree.algorithms.barycentric_wachspress import (
    barycentric_wachspress_weights,
)
from numba_celltree.algorithms.cohen_sutherland import cohen_sutherland_line_box_clip
from numba_celltree.algorithms.cyrus_beck import cyrus_beck_line_polygon_clip
from numba_celltree.algorithms.liang_barsky import liang_barsky_line_box_clip
from numba_celltree.algorithms.separating_axis import polygons_intersect
from numba_celltree.algorithms.sutherland_hodgman import (
    area_of_intersection,
    box_area_of_intersection,
)

__all__ = (
    "barycentric_triangle_weights",
    "barycentric_wachspress_weights",
    "cohen_sutherland_line_box_clip",
    "cyrus_beck_line_polygon_clip",
    "liang_barsky_line_box_clip",
    "polygons_intersect",
    "area_of_intersection",
    "box_area_of_intersection",
)



"""
Implementation based off the description in:
Skala, V. (1993). An efficient algorithm for line clipping by convex polygon.
Computers & Graphics, 17(4), 417-421.

Available at:
https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.302.5729&rep=rep1&type=pdf

Also available in:
Duc Huy Bui, 1999. Algorithms for Line Clipping and Their Complexity. PhD
Thesis.

Available at:
http://graphics.zcu.cz/files/DIS_1999_Bui_Duc_Huy.pdf
"""

from typing import Sequence, Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import Point, Vector
from numba_celltree.geometry_utils import (
    as_point,
    cross_product,
    dot_product,
    point_in_polygon_or_on_edge,
    to_point,
    to_vector,
)

NO_INTERSECTION = False, Point(np.nan, np.nan), Point(np.nan, np.nan)


@nb.njit(inline="always")
def compute_intersection(
    a: Point, s: Vector, v0: Point, v1: Point
) -> Tuple[bool, float]:
    # Computes intersection for parametrized vector
    # vector given by a & s
    # line given by polygon vertices v0 & v1
    # Due to ksi_eta check, k should never be 0 (parallel, possibly collinear)
    # Note: Polygon must be counter-clockwise
    si = to_vector(a, v0)
    n = Vector(-(v1.y - v0.y), (v1.x - v0.x))
    n_si = dot_product(n, si)
    k = dot_product(n, s)
    t = n_si / k
    if n_si > 0:
        return True, t  # Entering
    else:
        return False, t  # leaving


@nb.njit(inline="always")
def intersections(
    a: Point, s: Vector, poly: Sequence[Point], length: int, k: int, i0: int, i1: int
) -> Tuple[float, float]:
    # Return t's for parametrized vector
    # Note: polygon must be counter-clockwise

    # A single intersection found, could be entering, could be leaving
    v0 = as_point(poly[i0])
    v01 = as_point(poly[(i0 + 1) % length])
    v1 = as_point(poly[i1])
    v11 = as_point(poly[(i1 + 1) % length])
    _, t0 = compute_intersection(a, s, v0, v01)
    enters1, t1 = compute_intersection(a, s, v1, v11)
    if enters1:  # Swap them
        return t1, t0
    else:
        return t0, t1


@nb.njit(inline="always")
def overlap(ta: Point, tb: Point, t0: Point, t1: Point) -> bool:
    if ta > tb:
        ta, tb = tb, ta
    if t0 > t1:
        t0, t1 = t1, t0
    vector_overlap = max(0, min(tb, t1) - max(ta, t0))
    return vector_overlap > 0.0


@nb.njit(inline="always")
def collinear_case(a: Point, b: Point, v0: Point, v1: Point) -> Tuple[Point, Point]:
    # Project on the same axis (t), take inner values
    n = Vector(-(b.y - a.y), (b.x - a.x))
    ta = cross_product(n, a)
    tb = cross_product(n, b)
    t0 = cross_product(n, v0)
    t1 = cross_product(n, v1)

    if not overlap(ta, tb, t0, t1):
        return NO_INTERSECTION

    if t0 < ta:
        p0 = v0
    else:
        p0 = a

    if t1 > tb:
        p1 = v1
    else:
        p1 = b

    return True, p0, p1


# Too big to inline. Drives compilation time through the roof for no benefit.
@nb.njit(inline="never")
def cyrus_beck_line_polygon_clip(
    a: Point, b: Point, poly: Sequence[Point]
) -> Tuple[bool, Point, Point]:
    """
    In short, the basic idea:

    For a given segment s (a -> b), test which two edges of a convex polygon it
    can intersect. If it intersects, the vertices [v0, v1] of an edge are
    separated by the segment (s). If s separates, the cross products of a -> v0
    (ksi) and a -> v1 (eta) will point in opposing directions (ksi * eta < 0).
    If both are > 0 or both are < 0 (ksi * eta > 0), they fall on the on the
    same side of the line; they are parallel and possibly collinear if ksi *
    eta == 0.

    Once the number of intersections (k), and the possibly intersecting edges
    (i0, i1) have been identified, we can compute the intersections. This
    assumes the vertices of the polygons are ordered in counter-clockwise
    orientation. We can also tell whether a line is possibly entering or
    leaving the polygon by the sign of the dot product.

    A valid intersection falls on the domain of the parametrized segment:
    0 <= t <= 1.0
    """
    length = len(poly)
    s = to_vector(a, b)

    # Test whether points are identical
    if s.x == 0 and s.y == 0:
        return NO_INTERSECTION
    # Test whether line is fully enclosed in polygon
    a_inside = point_in_polygon_or_on_edge(a, poly)
    b_inside = point_in_polygon_or_on_edge(b, poly)
    if a_inside and b_inside:
        return True, a, b

    i0 = -1
    i1 = -1
    i = 0
    k = 0
    v = as_point(poly[i])
    ksi = cross_product(to_vector(a, v), s)

    while i < length and k < 2:
        v0 = as_point(poly[i])
        v1 = as_point(poly[(i + 1) % length])
        # Check if they can cross at all
        eta = cross_product(to_vector(a, v1), s)

        # Note; ksi * eta < 0 doesn't work as well
        if (ksi < 0.0) ^ (eta < 0.0):
            if k == 0:
                i0 = i
            else:
                i1 = i
            k += 1
        # Calculate the area of the triangle formed by a, b, v0
        # if zero, then points are collinear.
        elif (ksi == 0.0) and (eta == 0.0):
            return collinear_case(a, b, v0, v1)

        # Don't recompute ksi
        ksi = eta
        i += 1

    if k == 0:
        return NO_INTERSECTION

    # Gather the intersections, given half-planes
    t0, t1 = intersections(a, s, poly, length, k, i0, i1)

    # Deal with edge cases
    if t0 == t1:
        if a_inside and t1 != 0.0:
            t0 = 0.0
        elif b_inside and t0 != 1.0:
            t1 = 1.0
        else:
            return NO_INTERSECTION

    # Swap if necessary so that t0 is the smaller
    if t1 < t0:
        t0, t1 = t1, t0

    # Note:
    # t >= 0, not >
    # t1 <= 1, not <
    valid0 = t0 >= 0 and t0 < 1
    valid1 = t1 > 0 and t1 <= 1

    # Return only the intersections that are within the segment
    if valid0 and valid1:
        return True, to_point(t0, a, s), to_point(t1, a, s)
    elif valid0:
        return True, to_point(t0, a, s), b
    elif valid1:
        return True, a, to_point(t1, a, s)
    else:
        return NO_INTERSECTION
